# Copyright 2020-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import pathlib
import re
import subprocess
import sys
import zipfile
from textwrap import dedent
from typing import List
from unittest import mock
from unittest.mock import ANY, MagicMock, call, patch

import pytest
import yaml
from craft_cli import CraftError, EmitterMode, emit
from craft_providers import bases

from charmcraft import instrum, linters
from charmcraft.bases import get_host_as_base
from charmcraft.charm_builder import relativise
from charmcraft.config import load
from charmcraft.const import BUILD_DIRNAME
from charmcraft.models.charmcraft import Base, BasesConfiguration
from charmcraft.models.lint import LintResult
from charmcraft.package import (
    Builder,
    _subprocess_pack_charms,
    format_charm_file_name,
    launch_shell,
)
from charmcraft.providers import get_base_configuration
from charmcraft.utils import get_host_architecture


def get_builder(
    config,
    *,
    project_dir=None,
    force=False,
    debug=False,
    shell=False,
    shell_after=False,
    measure=None,
):
    if project_dir is None:
        project_dir = config.project.dirpath

    return Builder(
        config=config,
        debug=debug,
        force=force,
        shell=shell,
        shell_after=shell_after,
        measure=measure,
    )


@pytest.fixture()
def basic_project(tmp_path, monkeypatch, prepare_charmcraft_yaml, prepare_metadata_yaml):
    """Create a basic Charmcraft project."""
    build_dir = tmp_path / BUILD_DIRNAME
    build_dir.mkdir()

    # a lib dir
    lib_dir = tmp_path / "lib"
    lib_dir.mkdir()
    ops_lib_dir = lib_dir / "ops"
    ops_lib_dir.mkdir()
    ops_stuff = ops_lib_dir / "stuff.txt"
    ops_stuff.write_bytes(b"ops stuff")

    # simple source code
    src_dir = tmp_path / "src"
    src_dir.mkdir()
    charm_script = src_dir / "charm.py"
    charm_script.write_bytes(b"all the magic")
    charm_script.chmod(0o755)

    # the license file
    license = tmp_path / "LICENSE"
    license.write_text("license content")

    # other optional assets
    icon = tmp_path / "icon.svg"
    icon.write_text("icon content")

    # README
    readme = tmp_path / "README.md"
    readme.write_text("README content")

    # the config
    host_base = get_host_as_base()
    prepare_charmcraft_yaml(
        dedent(
            f"""
            type: charm
            bases:
              - name: {host_base.name}
                channel: "{host_base.channel}"
                architectures: {host_base.architectures!r}
            parts:  # just to avoid "default charm parts" sneaking in
              foo:
                plugin: nil
            """
        )
    )
    prepare_metadata_yaml(
        dedent(
            """\
            name: test-charm-name-from-metadata-yaml
            summary: test summary
            description: test description
            """
        ),
    )
    # paths are relative, make all tests to run in the project's directory
    monkeypatch.chdir(tmp_path)

    return tmp_path


@pytest.fixture()
def basic_project_builder(basic_project, prepare_charmcraft_yaml):
    def _basic_project_builder(bases_configs: List[BasesConfiguration], **builder_kwargs):
        charmcraft_yaml = dedent(
            """
            type: charm
            bases:
            """
        )

        for bases_config in bases_configs:
            charmcraft_yaml += "  - build-on:\n"
            for base in bases_config.build_on:
                charmcraft_yaml += (
                    f"    - name: {base.name!r}\n"
                    f"      channel: {base.channel!r}\n"
                    f"      architectures: {base.architectures!r}\n"
                )

            charmcraft_yaml += "    run-on:\n"
            for base in bases_config.run_on:
                charmcraft_yaml += (
                    f"    - name: {base.name!r}\n"
                    f"      channel: {base.channel!r}\n"
                    f"      architectures: {base.architectures!r}\n"
                )
        charmcraft_yaml += dedent(
            """
            parts:
              foo:
                plugin: nil
            """
        )

        prepare_charmcraft_yaml(charmcraft_yaml)

        config = load(basic_project)
        return get_builder(config, **builder_kwargs)

    return _basic_project_builder


@pytest.fixture()
def mock_capture_logs_from_instance():
    with patch("charmcraft.providers.capture_logs_from_instance") as mock_capture:
        yield mock_capture


@pytest.fixture()
def mock_launch_shell():
    with patch("charmcraft.package.launch_shell") as mock_shell:
        yield mock_shell


@pytest.fixture()
def mock_linters():
    with patch("charmcraft.linters") as mock_linters:
        mock_linters.analyze.return_value = []
        yield mock_linters


@pytest.fixture()
def mock_parts():
    with patch("charmcraft.parts") as mock_parts:
        yield mock_parts


@pytest.fixture(autouse=True)
def mock_provider(mock_instance, fake_provider):
    mock_provider = mock.Mock(wraps=fake_provider)
    with patch("charmcraft.providers.get_provider", return_value=mock_provider):
        yield mock_provider


@pytest.fixture()
def mock_ubuntu_buildd_base_configuration():
    with mock.patch("craft_providers.bases.ubuntu.BuilddBase", autospec=True) as mock_base_config:
        yield mock_base_config


@pytest.fixture()
def mock_centos_base_configuration():
    with mock.patch("craft_providers.bases.centos.CentOSBase", autospec=True) as mock_base_config:
        yield mock_base_config


@pytest.fixture()
def mock_instance_name():
    with mock.patch(
        "charmcraft.providers.get_instance_name", return_value="test-instance-name"
    ) as patched:
        yield patched


@pytest.fixture()
def mock_is_base_available():
    with mock.patch(
        "charmcraft.providers.is_base_available",
        return_value=(True, None),
    ) as mock_is_base_available:
        yield mock_is_base_available


# --- (real) build tests


def test_build_error_without_metadata_yaml(basic_project):
    """Validate error if trying to build project without metadata.yaml."""
    metadata = basic_project / "metadata.yaml"
    metadata.unlink()

    with pytest.raises(CraftError) as exc_info:
        config = load(basic_project)

        get_builder(config)

    assert str(exc_info.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - needs value in field 'name'
        - needs value in field 'summary'
        - needs value in field 'description'"""
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_with_charmcraft_yaml_destructive_mode(basic_project_builder, emitter, monkeypatch):
    host_base = get_host_as_base()
    builder = basic_project_builder(
        [BasesConfiguration(**{"build-on": [host_base], "run-on": [host_base]})],
        force=True,  # to ignore any linter issue
    )

    zipnames = builder.run(destructive_mode=True)

    host_arch = host_base.architectures[0]
    assert zipnames == [
        "test-charm-name-from-metadata-yaml_"
        f"{host_base.name}-{host_base.channel}-{host_arch}.charm"
    ]

    emitter.assert_debug("Building for 'bases[0]' as host matches 'build-on[0]'.")


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_with_charmcraft_yaml_managed_mode(
    basic_project_builder, emitter, monkeypatch, tmp_path
):
    monkeypatch.setenv("CHARMCRAFT_MANAGED_MODE", "1")
    host_base = get_host_as_base()
    builder = basic_project_builder(
        [BasesConfiguration(**{"build-on": [host_base], "run-on": [host_base]})],
        force=True,  # to ignore any linter issue
    )

    with patch("charmcraft.env.get_managed_environment_home_path", return_value=tmp_path / "root"):
        zipnames = builder.run()

    host_arch = host_base.architectures[0]
    assert zipnames == [
        "test-charm-name-from-metadata-yaml_"
        f"{host_base.name}-{host_base.channel}-{host_arch}.charm"
    ]

    emitter.assert_debug("Building for 'bases[0]' as host matches 'build-on[0]'.")


def test_build_checks_provider(basic_project, mock_provider, mock_capture_logs_from_instance):
    """Test cases for base-index parameter."""
    config = load(basic_project)
    builder = get_builder(config)

    try:
        builder.run()
    except CraftError:
        # 'No suitable 'build-on' environment...' error will be raised on some test platforms
        pass

    mock_provider.ensure_provider_is_available.assert_called_once()


def test_build_with_debug_no_error(
    basic_project_builder,
    mock_linters,
    mock_parts,
    mock_launch_shell,
):
    host_base = get_host_as_base()
    builder = basic_project_builder(
        [BasesConfiguration(**{"build-on": [host_base], "run-on": [host_base]})],
        debug=True,
    )

    charms = builder.run(destructive_mode=True)

    assert len(charms) == 1
    assert mock_launch_shell.mock_calls == []


def test_build_with_debug_with_error(
    basic_project_builder,
    mock_linters,
    mock_parts,
    mock_launch_shell,
):
    mock_parts.PartsLifecycle.return_value.run.side_effect = CraftError("fail")
    host_base = get_host_as_base()
    builder = basic_project_builder(
        [BasesConfiguration(**{"build-on": [host_base], "run-on": [host_base]})],
        debug=True,
    )

    with pytest.raises(CraftError):
        builder.run(destructive_mode=True)

    assert mock_launch_shell.mock_calls == [mock.call()]


def test_build_with_shell(basic_project_builder, mock_parts, mock_provider, mock_launch_shell):
    host_base = get_host_as_base()
    builder = basic_project_builder(
        [BasesConfiguration(**{"build-on": [host_base], "run-on": [host_base]})],
        shell=True,
    )

    charms = builder.run(destructive_mode=True)

    assert charms == []
    assert mock_launch_shell.mock_calls == [mock.call()]


def test_build_with_shell_after(
    basic_project_builder,
    mock_linters,
    mock_parts,
    mock_launch_shell,
):
    host_base = get_host_as_base()
    builder = basic_project_builder(
        [BasesConfiguration(**{"build-on": [host_base], "run-on": [host_base]})],
        shell_after=True,
    )

    charms = builder.run(destructive_mode=True)

    assert len(charms) == 1
    assert mock_launch_shell.mock_calls == [mock.call()]


def test_build_checks_provider_error(basic_project, mock_provider):
    """Test cases for base-index parameter."""
    mock_provider.ensure_provider_is_available.side_effect = RuntimeError("foo")
    config = load(basic_project)
    builder = get_builder(config)

    with pytest.raises(RuntimeError, match="foo"):
        builder.run()


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_multiple_with_charmcraft_yaml_destructive_mode(basic_project_builder, emitter):
    """Build multiple charms for multiple matching bases, skipping one unmatched config."""
    host_base = get_host_as_base()
    unmatched_base = Base(
        name="unmatched-name",
        channel="unmatched-channel",
        architectures=["unmatched-arch1"],
    )
    matched_cross_base = Base(
        name="cross-name",
        channel="cross-channel",
        architectures=["cross-arch1"],
    )
    builder = basic_project_builder(
        [
            BasesConfiguration(**{"build-on": [host_base], "run-on": [host_base]}),
            BasesConfiguration(**{"build-on": [unmatched_base], "run-on": [unmatched_base]}),
            BasesConfiguration(**{"build-on": [host_base], "run-on": [matched_cross_base]}),
        ],
        force=True,
    )

    zipnames = builder.run(destructive_mode=True)

    host_arch = host_base.architectures[0]
    assert zipnames == [
        "test-charm-name-from-metadata-yaml_"
        f"{host_base.name}-{host_base.channel}-{host_arch}.charm",
        "test-charm-name-from-metadata-yaml_cross-name-cross-channel-cross-arch1.charm",
    ]

    reason = f"name 'unmatched-name' does not match host {host_base.name!r}."
    emitter.assert_interactions(
        [
            call("debug", "Building for 'bases[0]' as host matches 'build-on[0]'."),
            call("progress", f"Skipping 'bases[1].build-on[0]': {reason}"),
            call(
                "progress",
                "No suitable 'build-on' environment found in 'bases[1]' configuration.",
                permanent=True,
            ),
            call("debug", "Building for 'bases[2]' as host matches 'build-on[0]'."),
        ]
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_multiple_with_charmcraft_yaml_managed_mode(
    basic_project_builder, monkeypatch, emitter, tmp_path
):
    """Build multiple charms for multiple matching bases, skipping one unmatched config."""
    host_base = get_host_as_base()
    unmatched_base = Base(
        name="unmatched-name",
        channel="unmatched-channel",
        architectures=["unmatched-arch1"],
    )
    matched_cross_base = Base(
        name="cross-name",
        channel="cross-channel",
        architectures=["cross-arch1"],
    )
    builder = basic_project_builder(
        [
            BasesConfiguration(**{"build-on": [host_base], "run-on": [host_base]}),
            BasesConfiguration(**{"build-on": [unmatched_base], "run-on": [unmatched_base]}),
            BasesConfiguration(**{"build-on": [host_base], "run-on": [matched_cross_base]}),
        ],
        force=True,
    )

    monkeypatch.setenv("CHARMCRAFT_MANAGED_MODE", "1")
    with patch("charmcraft.env.get_managed_environment_home_path", return_value=tmp_path / "root"):
        zipnames = builder.run()

    host_arch = host_base.architectures[0]
    assert zipnames == [
        "test-charm-name-from-metadata-yaml_"
        f"{host_base.name}-{host_base.channel}-{host_arch}.charm",
        "test-charm-name-from-metadata-yaml_cross-name-cross-channel-cross-arch1.charm",
    ]

    reason = f"name 'unmatched-name' does not match host {host_base.name!r}."
    emitter.assert_interactions(
        [
            call("debug", "Building for 'bases[0]' as host matches 'build-on[0]'."),
            call("progress", f"Skipping 'bases[1].build-on[0]': {reason}"),
            call(
                "progress",
                "No suitable 'build-on' environment found in 'bases[1]' configuration.",
                permanent=True,
            ),
            call("debug", "Building for 'bases[2]' as host matches 'build-on[0]'."),
        ]
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml_template", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: ubuntu
                    channel: "18.04"
                    architectures: {arch}
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
    ],
)
def test_build_project_is_cwd(
    basic_project,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml_template,
    metadata_yaml,
    emitter,
    mock_capture_logs_from_instance,
    mock_instance,
    mock_provider,
    mock_instance_name,
    mock_ubuntu_buildd_base_configuration,
    mock_is_base_available,
    mocker,
):
    """Test cases for base-index parameter."""
    emit.set_mode(EmitterMode.BRIEF)
    host_base = get_host_as_base()
    host_arch = host_base.architectures[0]
    prepare_charmcraft_yaml(charmcraft_yaml_template.format(arch=host_base.architectures))
    prepare_metadata_yaml(metadata_yaml)

    config = load(basic_project)
    project_managed_path = pathlib.Path("/root/project")
    builder = get_builder(config)
    base_configuration = get_base_configuration(
        alias=bases.ubuntu.BuilddBaseAlias.BIONIC, instance_name=mock_instance_name()
    )
    mock_base_get_base_configuration = mocker.patch("charmcraft.providers.get_base_configuration")
    mock_base_get_base_configuration.return_value = base_configuration

    zipnames = builder.run([0])

    assert zipnames == [
        f"test-charm-name-from-metadata-yaml_ubuntu-18.04-{host_arch}.charm",
    ]
    assert mock_provider.mock_calls == [
        call.is_provider_installed(),
        call.ensure_provider_is_available(),
        call.launched_environment(
            project_name="test-charm-name-from-metadata-yaml",
            project_path=basic_project,
            base_configuration=base_configuration,
            instance_name=mock_instance_name(),
            allow_unstable=False,
        ),
    ]
    assert mock_instance.mock_calls == [
        call.mount(host_source=basic_project, target=project_managed_path),
        call.execute_run(
            ["charmcraft", "pack", "--bases-index", "0", "--verbosity=brief"],
            check=True,
            cwd=project_managed_path,
        ),
    ]
    assert mock_is_base_available.mock_calls == [
        call.is_base_available(
            Base(name="ubuntu", channel="18.04", architectures=[host_arch]),
        )
    ]


@pytest.mark.parametrize(
    ("charmcraft_yaml_template", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: ubuntu
                    channel: "18.04"
                    architectures: {arch}
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
    ],
)
def test_build_project_is_not_cwd(
    basic_project,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml_template,
    metadata_yaml,
    mock_capture_logs_from_instance,
    mock_instance,
    mock_provider,
    monkeypatch,
    mock_instance_name,
    mock_ubuntu_buildd_base_configuration,
    mock_is_base_available,
    mocker,
):
    """Test cases for base-index parameter."""
    emit.set_mode(EmitterMode.BRIEF)
    host_base = get_host_as_base()
    host_arch = host_base.architectures[0]
    prepare_charmcraft_yaml(charmcraft_yaml_template.format(arch=host_base.architectures))
    prepare_metadata_yaml(metadata_yaml)

    config = load(basic_project)
    builder = get_builder(config)
    base_configuration = get_base_configuration(
        alias=bases.ubuntu.BuilddBaseAlias.BIONIC, instance_name=mock_instance_name()
    )
    mock_base_get_base_configuration = mocker.patch("charmcraft.providers.get_base_configuration")
    mock_base_get_base_configuration.return_value = base_configuration

    monkeypatch.chdir("/")  # make the working directory NOT the project's one
    zipnames = builder.run([0])

    assert zipnames == [
        f"test-charm-name-from-metadata-yaml_ubuntu-18.04-{host_arch}.charm",
    ]
    assert mock_provider.mock_calls == [
        call.is_provider_installed(),
        call.ensure_provider_is_available(),
        call.launched_environment(
            project_name="test-charm-name-from-metadata-yaml",
            project_path=basic_project,
            base_configuration=base_configuration,
            instance_name=mock_instance_name(),
            allow_unstable=False,
        ),
    ]
    assert mock_instance.mock_calls == [
        call.mount(host_source=basic_project, target=pathlib.Path("/root/project")),
        call.execute_run(
            ["charmcraft", "pack", "--bases-index", "0", "--verbosity=brief"],
            check=True,
            cwd=pathlib.Path("/root"),
        ),
        call.pull_file(
            source=pathlib.Path("/root") / zipnames[0],
            destination=pathlib.Path.cwd() / zipnames[0],
        ),
    ]
    assert mock_is_base_available.mock_calls == [
        call.is_base_available(
            Base(name="ubuntu", channel="18.04", architectures=[host_arch]),
        )
    ]


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
@pytest.mark.parametrize(
    ("mode", "cmd_flags"),
    [
        (EmitterMode.VERBOSE, ["--verbosity=verbose"]),
        (EmitterMode.QUIET, ["--verbosity=quiet"]),
        (EmitterMode.DEBUG, ["--verbosity=debug"]),
        (EmitterMode.TRACE, ["--verbosity=trace"]),
        (EmitterMode.BRIEF, ["--verbosity=brief"]),
    ],
)
@pytest.mark.parametrize(
    ("charmcraft_yaml_template", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: ubuntu
                    channel: "18.04"
                    architectures: {arch}
                  - name: ubuntu
                    channel: "20.04"
                    architectures: {arch}
                  - name: centos
                    channel: "7"
                    architectures: {arch}
                  - name: ubuntu
                    channel: "unsupported-channel"
                    architectures: {arch}
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
    ],
)
def test_build_bases_index_scenarios_provider(
    mode,
    cmd_flags,
    basic_project,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml_template,
    metadata_yaml,
    emitter,
    mock_capture_logs_from_instance,
    mock_instance,
    mock_provider,
    mock_instance_name,
    mock_ubuntu_buildd_base_configuration,
    mock_centos_base_configuration,
    mock_is_base_available,
    mocker,
):
    """Test cases for base-index parameter."""
    emit.set_mode(mode)
    host_base = get_host_as_base()
    host_arch = host_base.architectures[0]
    project_managed_path = pathlib.Path("/root/project")
    prepare_charmcraft_yaml(charmcraft_yaml_template.format(arch=host_base.architectures))
    config = load(basic_project)
    builder = get_builder(config)
    base_bionic_configuration = get_base_configuration(
        alias=bases.ubuntu.BuilddBaseAlias.BIONIC, instance_name=mock_instance_name()
    )
    base_focal_configuration = get_base_configuration(
        alias=bases.ubuntu.BuilddBaseAlias.FOCAL, instance_name=mock_instance_name()
    )
    base_centos_configuration = get_base_configuration(
        alias=bases.centos.CentOSBaseAlias.SEVEN, instance_name=mock_instance_name()
    )
    mock_base_get_base_configuration = mocker.patch("charmcraft.providers.get_base_configuration")
    mock_base_get_base_configuration.side_effect = [
        base_bionic_configuration,
        base_focal_configuration,
        base_centos_configuration,
    ]

    zipnames = builder.run([0])
    assert zipnames == [
        f"test-charm-name-from-metadata-yaml_ubuntu-18.04-{host_arch}.charm",
    ]

    assert mock_provider.mock_calls == [
        call.is_provider_installed(),
        call.ensure_provider_is_available(),
        call.launched_environment(
            project_name="test-charm-name-from-metadata-yaml",
            project_path=basic_project,
            base_configuration=base_bionic_configuration,
            instance_name=mock_instance_name(),
            allow_unstable=False,
        ),
    ]
    assert mock_instance.mock_calls == [
        call.mount(host_source=basic_project, target=project_managed_path),
        call.execute_run(
            ["charmcraft", "pack", "--bases-index", "0", *cmd_flags],
            check=True,
            cwd=project_managed_path,
        ),
    ]
    assert mock_is_base_available.mock_calls == [
        call(Base(name="ubuntu", channel="18.04", architectures=[host_arch]))
    ]
    emitter.assert_progress(
        "Launching environment to pack for base "
        "name='ubuntu' channel='18.04' architectures=['amd64'] "
        "(may take a while the first time but it's reusable)"
    )
    emitter.assert_progress("Packing the charm")
    mock_provider.reset_mock()
    mock_instance.reset_mock()
    mock_is_base_available.reset_mock()
    mock_base_get_base_configuration.reset_mock()
    mock_base_get_base_configuration.side_effect = [
        base_bionic_configuration,
        base_focal_configuration,
        base_centos_configuration,
    ]

    zipnames = builder.run([1])
    assert zipnames == [
        f"test-charm-name-from-metadata-yaml_ubuntu-20.04-{host_arch}.charm",
    ]
    assert mock_provider.mock_calls == [
        call.is_provider_installed(),
        call.ensure_provider_is_available(),
        call.launched_environment(
            project_name="test-charm-name-from-metadata-yaml",
            project_path=basic_project,
            base_configuration=base_focal_configuration,
            instance_name=mock_instance_name(),
            allow_unstable=False,
        ),
    ]
    assert mock_instance.mock_calls == [
        call.mount(host_source=basic_project, target=project_managed_path),
        call.execute_run(
            ["charmcraft", "pack", "--bases-index", "1", *cmd_flags],
            check=True,
            cwd=project_managed_path,
        ),
    ]
    assert mock_is_base_available.mock_calls == [
        call(Base(name="ubuntu", channel="20.04", architectures=[host_arch]))
    ]
    mock_provider.reset_mock()
    mock_instance.reset_mock()
    mock_is_base_available.reset_mock()
    mock_base_get_base_configuration.reset_mock()
    mock_base_get_base_configuration.side_effect = [
        base_bionic_configuration,
        base_focal_configuration,
        base_centos_configuration,
    ]

    zipnames = builder.run([0, 1])
    assert zipnames == [
        f"test-charm-name-from-metadata-yaml_ubuntu-18.04-{host_arch}.charm",
        f"test-charm-name-from-metadata-yaml_ubuntu-20.04-{host_arch}.charm",
    ]
    assert mock_provider.mock_calls == [
        call.is_provider_installed(),
        call.ensure_provider_is_available(),
        call.launched_environment(
            project_name="test-charm-name-from-metadata-yaml",
            project_path=basic_project,
            base_configuration=base_bionic_configuration,
            instance_name=mock_instance_name(),
            allow_unstable=False,
        ),
        call.launched_environment(
            project_name="test-charm-name-from-metadata-yaml",
            project_path=basic_project,
            base_configuration=base_focal_configuration,
            instance_name=mock_instance_name(),
            allow_unstable=False,
        ),
    ]
    assert mock_instance.mock_calls == [
        call.mount(host_source=basic_project, target=project_managed_path),
        call.execute_run(
            ["charmcraft", "pack", "--bases-index", "0", *cmd_flags],
            check=True,
            cwd=project_managed_path,
        ),
        call.mount(host_source=basic_project, target=project_managed_path),
        call.execute_run(
            ["charmcraft", "pack", "--bases-index", "1", *cmd_flags],
            check=True,
            cwd=project_managed_path,
        ),
    ]
    assert mock_is_base_available.mock_calls == [
        call(Base(name="ubuntu", channel="18.04", architectures=[host_arch])),
        call(Base(name="ubuntu", channel="20.04", architectures=[host_arch])),
    ]

    mock_provider.reset_mock()
    mock_instance.reset_mock()
    mock_is_base_available.reset_mock()
    mock_base_get_base_configuration.reset_mock()
    mock_base_get_base_configuration.side_effect = [
        base_focal_configuration,
        base_centos_configuration,
    ]

    zipnames = builder.run([1, 2])
    assert zipnames == [
        f"test-charm-name-from-metadata-yaml_ubuntu-20.04-{host_arch}.charm",
        f"test-charm-name-from-metadata-yaml_centos-7-{host_arch}.charm",
    ]
    assert mock_provider.mock_calls == [
        call.is_provider_installed(),
        call.ensure_provider_is_available(),
        call.launched_environment(
            project_name="test-charm-name-from-metadata-yaml",
            project_path=basic_project,
            base_configuration=base_focal_configuration,
            instance_name=mock_instance_name(),
            allow_unstable=False,
        ),
        call.launched_environment(
            project_name="test-charm-name-from-metadata-yaml",
            project_path=basic_project,
            base_configuration=base_centos_configuration,
            instance_name=mock_instance_name(),
            allow_unstable=True,
        ),
    ]
    assert mock_instance.mock_calls == [
        call.mount(host_source=basic_project, target=project_managed_path),
        call.execute_run(
            ["charmcraft", "pack", "--bases-index", "1", *cmd_flags],
            check=True,
            cwd=project_managed_path,
        ),
        call.mount(host_source=basic_project, target=project_managed_path),
        call.execute_run(
            ["charmcraft", "pack", "--bases-index", "2", *cmd_flags],
            check=True,
            cwd=project_managed_path,
        ),
    ]
    assert mock_is_base_available.mock_calls == [
        call(Base(name="ubuntu", channel="20.04", architectures=[host_arch])),
        call(Base(name="centos", channel="7", architectures=[host_arch])),
    ]
    mock_provider.reset_mock()
    mock_instance.reset_mock()
    mock_is_base_available.reset_mock()
    mock_base_get_base_configuration.reset_mock()
    mock_base_get_base_configuration.side_effect = [
        base_bionic_configuration,
        base_focal_configuration,
        base_centos_configuration,
    ]

    with pytest.raises(
        CraftError,
        match=r"No suitable 'build-on' environment found in any 'bases' configuration.",
    ):
        builder.run([4])

    mock_provider.reset_mock()
    mock_instance.reset_mock()
    mock_is_base_available.reset_mock()
    mock_base_get_base_configuration.reset_mock()
    mock_base_get_base_configuration.side_effect = [
        base_bionic_configuration,
        base_focal_configuration,
    ]

    expected_msg = re.escape("Failed to build charm for bases index '0'.")
    with pytest.raises(
        CraftError,
        match=expected_msg,
    ):
        mock_instance.execute_run.side_effect = subprocess.CalledProcessError(
            -1,
            ["charmcraft", "pack", "..."],
            "some output",
            "some error",
        )
        builder.run([0])

    assert mock_instance.mock_calls == [
        call.mount(host_source=basic_project, target=project_managed_path),
        call.execute_run(
            ["charmcraft", "pack", "--bases-index", "0", *cmd_flags],
            check=True,
            cwd=project_managed_path,
        ),
    ]
    # it was called seven times, for success and errors
    assert mock_capture_logs_from_instance.mock_calls == [call(mock_instance)] * 7


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_bases_index_scenarios_managed_mode(basic_project_builder, monkeypatch, tmp_path):
    """Test cases for base-index parameter."""
    host_base = get_host_as_base()
    host_arch = host_base.architectures[0]
    host_base = get_host_as_base()
    unmatched_base = Base(
        name="unmatched-name",
        channel="unmatched-channel",
        architectures=["unmatched-arch1"],
    )
    matched_cross_base = Base(
        name="cross-name",
        channel="cross-channel",
        architectures=["cross-arch1"],
    )
    builder = basic_project_builder(
        [
            BasesConfiguration(**{"build-on": [host_base], "run-on": [host_base]}),
            BasesConfiguration(**{"build-on": [unmatched_base], "run-on": [unmatched_base]}),
            BasesConfiguration(**{"build-on": [host_base], "run-on": [matched_cross_base]}),
        ],
        force=True,
    )

    monkeypatch.setenv("CHARMCRAFT_MANAGED_MODE", "1")
    with patch("charmcraft.env.get_managed_environment_home_path", return_value=tmp_path / "root"):
        zipnames = builder.run([0])
    assert zipnames == [
        "test-charm-name-from-metadata-yaml_"
        f"{host_base.name}-{host_base.channel}-{host_arch}.charm",
    ]

    with pytest.raises(
        CraftError,
        match=r"No suitable 'build-on' environment found in any 'bases' configuration.",
    ):
        builder.run([1])

    with patch("charmcraft.env.get_managed_environment_home_path", return_value=tmp_path / "root"):
        zipnames = builder.run([2])
    assert zipnames == [
        "test-charm-name-from-metadata-yaml_cross-name-cross-channel-cross-arch1.charm",
    ]


@patch(
    "charmcraft.bases.get_host_as_base",
    return_value=Base(name="xname", channel="xchannel", architectures=["xarch"]),
)
@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: unmatched-name
                    channel: xchannel
                    architectures: [xarch]
                  - name: xname
                    channel: unmatched-channel
                    architectures: [xarch]
                  - name: xname
                    channel: xchannel
                    architectures: [unmatched-arch1, unmatched-arch2]
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
    ],
)
def test_build_error_no_match_with_charmcraft_yaml(
    mock_host_base,
    basic_project,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml,
    metadata_yaml,
    monkeypatch,
    emitter,
):
    """Error when no charms are buildable with host base, verifying each mismatched reason."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(basic_project)
    builder = get_builder(config)

    # Managed bases build.
    monkeypatch.setenv("CHARMCRAFT_MANAGED_MODE", "1")
    with pytest.raises(
        CraftError,
        match=r"No suitable 'build-on' environment found in any 'bases' configuration.",
    ):
        builder.run()

    emitter.assert_interactions(
        [
            call(
                "progress",
                "Skipping 'bases[0].build-on[0]': "
                "name 'unmatched-name' does not match host 'xname'.",
            ),
            call(
                "progress",
                "No suitable 'build-on' environment found in 'bases[0]' configuration.",
                permanent=True,
            ),
            call(
                "progress",
                "Skipping 'bases[1].build-on[0]': "
                "channel 'unmatched-channel' does not match host 'xchannel'.",
            ),
            call(
                "progress",
                "No suitable 'build-on' environment found in 'bases[1]' configuration.",
                permanent=True,
            ),
            call(
                "progress",
                "Skipping 'bases[2].build-on[0]': "
                "host architecture 'xarch' not in base architectures "
                "['unmatched-arch1', 'unmatched-arch2'].",
            ),
            call(
                "progress",
                "No suitable 'build-on' environment found in 'bases[2]' configuration.",
                permanent=True,
            ),
        ]
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
@pytest.mark.parametrize(
    ("builder_flag", "cmd_flag"),
    [
        ("debug", "--debug"),
        ("shell", "--shell"),
        ("shell_after", "--shell-after"),
        ("force", "--force"),
    ],
)
def test_build_arguments_managed_charmcraft_simples(
    builder_flag,
    cmd_flag,
    mock_capture_logs_from_instance,
    mock_instance,
    basic_project_builder,
):
    """Check that the command to run charmcraft inside the environment is properly built."""
    emit.set_mode(EmitterMode.BRIEF)
    host_base = Base(name="ubuntu", channel="18.04", architectures=[get_host_architecture()])
    bases_config = [BasesConfiguration(**{"build-on": [host_base], "run-on": [host_base]})]
    project_managed_path = pathlib.Path("/root/project")

    kwargs = {builder_flag: True}
    builder = basic_project_builder(bases_config, **kwargs)
    builder.pack_charm_in_instance(
        build_on=bases_config[0].build_on[0],
        bases_index=0,
        build_on_index=0,
    )
    expected_cmd = ["charmcraft", "pack", "--bases-index", "0", "--verbosity=brief", cmd_flag]
    assert mock_instance.mock_calls == [
        call.mount(host_source=builder.config.project.dirpath, target=project_managed_path),
        call.execute_run(expected_cmd, check=True, cwd=project_managed_path),
    ]


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_arguments_managed_charmcraft_measure(
    mock_capture_logs_from_instance,
    mock_instance,
    basic_project_builder,
    tmp_path,
):
    """Check that the command to run charmcraft inside the environment is properly built."""
    emit.set_mode(EmitterMode.BRIEF)
    host_base = Base(name="ubuntu", channel="18.04", architectures=[get_host_architecture()])
    bases_config = [BasesConfiguration(**{"build-on": [host_base], "run-on": [host_base]})]
    project_managed_path = pathlib.Path("/root/project")

    # fake a dumped measure to be pulled from the instance
    fake_local_m = tmp_path / "local.json"
    instrum._Measurements().dump(fake_local_m)

    # specially patch the context manager
    fake_inst_m = tmp_path / "inst.json"
    mock_instance.temporarily_pull_file = MagicMock()
    mock_instance.temporarily_pull_file.return_value = fake_local_m

    builder = basic_project_builder(bases_config, measure=tmp_path)
    with patch("charmcraft.env.get_managed_environment_metrics_path", return_value=fake_inst_m):
        builder.pack_charm_in_instance(
            build_on=bases_config[0].build_on[0],
            bases_index=0,
            build_on_index=0,
        )
    cmd_flag = f"--measure={str(fake_inst_m)}"
    expected_cmd = ["charmcraft", "pack", "--bases-index", "0", "--verbosity=brief", cmd_flag]
    assert mock_instance.mock_calls == [
        call.mount(host_source=builder.config.project.dirpath, target=project_managed_path),
        call.execute_run(expected_cmd, check=True, cwd=project_managed_path),
        call.temporarily_pull_file(fake_inst_m),
    ]


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_package_tree_structure(tmp_path, config):
    """The zip file is properly built internally."""
    # the metadata
    metadata_data = {"name": "test-charm-name-from-metadata-yaml"}
    metadata_file = tmp_path / "metadata.yaml"
    with metadata_file.open("wt", encoding="ascii") as fh:
        yaml.dump(metadata_data, fh)

    # create some dirs and files! a couple of files outside, and the dir we'll zip...
    file_outside_1 = tmp_path / "file_outside_1"
    with file_outside_1.open("wb") as fh:
        fh.write(b"content_out_1")
    file_outside_2 = tmp_path / "file_outside_2"
    with file_outside_2.open("wb") as fh:
        fh.write(b"content_out_2")
    to_be_zipped_dir = tmp_path / BUILD_DIRNAME
    to_be_zipped_dir.mkdir()

    # ...also outside a dir with a file...
    dir_outside = tmp_path / "extdir"
    dir_outside.mkdir()
    file_ext = dir_outside / "file_ext"
    with file_ext.open("wb") as fh:
        fh.write(b"external file")

    # ...then another file inside, and another dir...
    file_inside = to_be_zipped_dir / "file_inside"
    with file_inside.open("wb") as fh:
        fh.write(b"content_in")
    dir_inside = to_be_zipped_dir / "somedir"
    dir_inside.mkdir()

    # ...also inside, a link to the external dir...
    dir_linked_inside = to_be_zipped_dir / "linkeddir"
    dir_linked_inside.symlink_to(dir_outside)

    # ...and finally another real file, and two symlinks
    file_deep_1 = dir_inside / "file_deep_1"
    with file_deep_1.open("wb") as fh:
        fh.write(b"content_deep")
    file_deep_2 = dir_inside / "file_deep_2"
    file_deep_2.symlink_to(file_inside)
    file_deep_3 = dir_inside / "file_deep_3"
    file_deep_3.symlink_to(file_outside_1)

    # zip it
    bases_config = BasesConfiguration(
        **{
            "build-on": [],
            "run-on": [Base(name="xname", channel="xchannel", architectures=["xarch1"])],
        }
    )
    builder = get_builder(config)
    zipname = builder.handle_package(to_be_zipped_dir, bases_config)

    # check the stuff outside is not in the zip, the stuff inside is zipped (with
    # contents!), and all relative to build dir
    zf = zipfile.ZipFile(zipname)
    assert "file_outside_1" not in [x.filename for x in zf.infolist()]
    assert "file_outside_2" not in [x.filename for x in zf.infolist()]
    assert zf.read("file_inside") == b"content_in"
    assert zf.read("somedir/file_deep_1") == b"content_deep"  # own
    assert zf.read("somedir/file_deep_2") == b"content_in"  # from file inside
    assert zf.read("somedir/file_deep_3") == b"content_out_1"  # from file outside 1
    assert zf.read("linkeddir/file_ext") == b"external file"  # from file in the outside linked dir


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml", "expected_zipname"),
    [
        [
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
            "test-charm-name-from-metadata-yaml_xname-xchannel-xarch1.charm",
        ],
        [
            dedent(
                """\
                name: test-charm-name-from-charmcraft-yaml
                type: charm
                summary: test summary
                description: test description
                """
            ),
            None,
            "test-charm-name-from-charmcraft-yaml_xname-xchannel-xarch1.charm",
        ],
    ],
)
def test_build_package_name(
    tmp_path,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml,
    metadata_yaml,
    expected_zipname,
):
    """The zip file name comes from the config."""
    to_be_zipped_dir = tmp_path / BUILD_DIRNAME
    to_be_zipped_dir.mkdir()

    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    # zip it
    bases_config = BasesConfiguration(
        **{
            "build-on": [],
            "run-on": [Base(name="xname", channel="xchannel", architectures=["xarch1"])],
        }
    )

    config = load(tmp_path)
    builder = get_builder(config)
    zipname = builder.handle_package(to_be_zipped_dir, bases_config)

    assert zipname == expected_zipname


@pytest.mark.parametrize(
    ("charmcraft_yaml_template", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - build-on:
                      - name: {base_name}
                        channel: "{base_channel}"
                    run-on:
                      - name: {base_name}
                        channel: "{base_channel}"
                parts:
                  charm:
                    charm-entrypoint: my_entrypoint.py
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
    ],
)
def test_build_postlifecycle_validation_is_properly_called(
    basic_project,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml_template,
    metadata_yaml,
    monkeypatch,
):
    """Check how the entrypoint validation helper is called."""
    host_base = get_host_as_base()
    prepare_charmcraft_yaml(
        charmcraft_yaml_template.format(base_name=host_base.name, base_channel=host_base.channel)
    )
    config = load(basic_project)
    builder = get_builder(config)

    entrypoint = basic_project / "my_entrypoint.py"
    entrypoint.touch(mode=0o700)

    monkeypatch.setenv("CHARMCRAFT_MANAGED_MODE", "1")
    with patch("charmcraft.parts.PartsLifecycle") as mock_lifecycle:
        mock_lifecycle.return_value = mock_lifecycle_instance = MagicMock()
        mock_lifecycle_instance.prime_dir = basic_project
        mock_lifecycle_instance.run().return_value = None
        with patch("charmcraft.linters.analyze"):
            with patch.object(Builder, "show_linting_results"):
                builder.run([0])


@pytest.mark.parametrize(
    ("charmcraft_yaml_template", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - build-on:
                      - name: {base_name}
                        channel: "{base_channel}"
                    run-on:
                      - name: {base_name}
                        channel: "{base_channel}"

                parts:
                  charm:
                    charm-entrypoint: src/charm.py
                    charm-python-packages: ["foo", "bar"]
                    charm-binary-python-packages: ["baz"]
                    charm-requirements: ["reqs.txt"]
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
    ],
)
def test_build_part_from_config(
    basic_project,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml_template,
    metadata_yaml,
    monkeypatch,
):
    """Check that the "parts" are passed to the lifecycle correctly."""
    host_base = get_host_as_base()
    prepare_charmcraft_yaml(
        charmcraft_yaml_template.format(base_name=host_base.name, base_channel=host_base.channel)
    )
    prepare_metadata_yaml(metadata_yaml)

    reqs_file = basic_project / "reqs.txt"
    reqs_file.write_text("somedep")
    config = load(basic_project)
    builder = get_builder(config, force=True)

    monkeypatch.setenv("CHARMCRAFT_MANAGED_MODE", "1")
    with patch("charmcraft.parts.PartsLifecycle", autospec=True) as mock_lifecycle:
        mock_lifecycle.side_effect = SystemExit()
        with pytest.raises(SystemExit):
            builder.run([0])
    mock_lifecycle.assert_has_calls(
        [
            call(
                {
                    "charm": {
                        "plugin": "charm",
                        "prime": ANY,
                        "charm-entrypoint": "src/charm.py",
                        "charm-python-packages": ["foo", "bar"],
                        "charm-binary-python-packages": ["baz"],
                        "source": str(basic_project),
                        "charm-requirements": ["reqs.txt"],
                        "charm-strict-dependencies": False,
                    }
                },
                work_dir=pathlib.Path("/root"),
                project_dir=basic_project,
                project_name="test-charm-name-from-metadata-yaml",
                ignore_local_sources=["*.charm"],
            )
        ]
    )
    assert set(mock_lifecycle.call_args_list[0][0][0]["charm"]["prime"]) == {
        "src",
        "venv",
        "hooks",
        "dispatch",
        "LICENSE",
        "README.md",
        "icon.svg",
        "lib",
        "metadata.yaml",
    }


@pytest.mark.parametrize(
    ("charmcraft_yaml_template", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - build-on:
                      - name: {base_name}
                        channel: "{base_channel}"
                    run-on:
                      - name: {base_name}
                        channel: "{base_channel}"

                parts:
                  charm:
                    charm-entrypoint: src/charm.py
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
    ],
)
def test_build_part_include_venv_pydeps(
    basic_project,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml_template,
    metadata_yaml,
    monkeypatch,
):
    """Include the venv directory even if only charmlib python dependencies exist."""
    host_base = get_host_as_base()
    prepare_charmcraft_yaml(
        charmcraft_yaml_template.format(base_name=host_base.name, base_channel=host_base.channel)
    )
    prepare_metadata_yaml(metadata_yaml)

    charmlib = basic_project / "lib" / "charms" / "somecharm" / "v1" / "somelib.py"
    charmlib.parent.mkdir(parents=True)
    charmlib.write_text(
        dedent(
            """
            LIBID = "asdasds"
            LIBAPI = 1
            LIBPATCH = 1
            PYDEPS = ["foo"]
            """
        )
    )
    config = load(basic_project)
    builder = get_builder(config, force=True)

    monkeypatch.setenv("CHARMCRAFT_MANAGED_MODE", "1")
    with patch("charmcraft.parts.PartsLifecycle", autospec=True) as mock_lifecycle:
        mock_lifecycle.side_effect = SystemExit()
        with pytest.raises(SystemExit):
            builder.run([0])
    mock_lifecycle.assert_has_calls(
        [
            call(
                {
                    "charm": {
                        "plugin": "charm",
                        "prime": ANY,
                        "charm-entrypoint": "src/charm.py",
                        "charm-python-packages": [],
                        "charm-binary-python-packages": [],
                        "source": str(basic_project),
                        "charm-requirements": [],
                        "charm-strict-dependencies": False,
                    }
                },
                work_dir=pathlib.Path("/root"),
                project_dir=basic_project,
                project_name="test-charm-name-from-metadata-yaml",
                ignore_local_sources=["*.charm"],
            )
        ]
    )

    assert set(mock_lifecycle.call_args_list[0][0][0]["charm"]["prime"]) == {
        "src",
        "venv",
        "hooks",
        "dispatch",
        "LICENSE",
        "README.md",
        "icon.svg",
        "lib",
        "metadata.yaml",
    }


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_build_using_linters_attributes(basic_project_builder, monkeypatch, tmp_path):
    """Generic use of linters, pass them ok to their processor and save them in the manifest."""
    host_base = get_host_as_base()
    builder = basic_project_builder(
        [BasesConfiguration(**{"build-on": [host_base], "run-on": [host_base]})],
        force=True,  # to ignore any linter issue
    )

    # the results from the analyzer
    linting_results = [
        linters.CheckResult(
            name="check-name-1",
            check_type=linters.CheckType.ATTRIBUTE,
            url="url",
            text="text",
            result="check-result-1",
        ),
        linters.CheckResult(
            name="check-name-2",
            check_type=linters.CheckType.ATTRIBUTE,
            url="url",
            text="text",
            result=LintResult.IGNORED,
        ),
    ]

    monkeypatch.setenv("CHARMCRAFT_MANAGED_MODE", "1")
    with patch("charmcraft.env.get_managed_environment_home_path", return_value=tmp_path / "root"):
        with patch("charmcraft.linters.analyze") as mock_analyze:
            with patch.object(Builder, "show_linting_results") as mock_show_lint:
                mock_analyze.return_value = linting_results
                zipnames = builder.run()

    # check the analyze and processing functions were called properly
    mock_analyze.assert_called_with(builder.config, tmp_path / "root" / "prime")
    mock_show_lint.assert_called_with(linting_results)

    # the manifest should have all the results (including the ignored one)
    zf = zipfile.ZipFile(zipnames[0])
    manifest = yaml.safe_load(zf.read("manifest.yaml"))
    expected = {
        "attributes": [
            {"name": "check-name-1", "result": "check-result-1"},
            {"name": "check-name-2", "result": "ignored"},
        ]
    }
    assert manifest["analysis"] == expected


def test_show_linters_attributes(basic_project, emitter, config):
    """Show the linting results, only attributes, one ignored."""
    builder = get_builder(config)

    # fake results from the analyzer
    linting_results = [
        linters.CheckResult(
            name="check-name-1",
            check_type=linters.CheckType.ATTRIBUTE,
            url="url",
            text="text",
            result="check-result-1",
        ),
        linters.CheckResult(
            name="check-name-2",
            check_type=linters.CheckType.ATTRIBUTE,
            url="url",
            text="text",
            result=LintResult.IGNORED,
        ),
    ]

    builder.show_linting_results(linting_results)

    expected = "Check result: check-name-1 [attribute] check-result-1 (text; see more at url)."
    emitter.assert_verbose(expected)


def test_show_linters_lint_warnings(basic_project, emitter, config):
    """Show the linting results, some warnings."""
    builder = get_builder(config)

    # fake result from the analyzer
    linting_results = [
        linters.CheckResult(
            name="check-name",
            check_type=linters.CheckType.LINT,
            url="check-url",
            text="Some text",
            result=LintResult.WARNINGS,
        ),
    ]

    builder.show_linting_results(linting_results)

    emitter.assert_interactions(
        [
            call("progress", "Lint Warnings:", permanent=True),
            call("progress", "- check-name: Some text (check-url)", permanent=True),
        ]
    )


def test_show_linters_lint_errors_normal(basic_project, emitter, config):
    """Show the linting results, have errors."""
    builder = get_builder(config)

    # fake result from the analyzer
    linting_results = [
        linters.CheckResult(
            name="check-name",
            check_type=linters.CheckType.LINT,
            url="check-url",
            text="Some text",
            result=LintResult.ERRORS,
        ),
    ]

    with pytest.raises(CraftError) as cm:
        builder.show_linting_results(linting_results)
    exc = cm.value
    assert str(exc) == "Aborting due to lint errors (use --force to override)."
    assert exc.retcode == 2

    emitter.assert_interactions(
        [
            call("progress", "Lint Errors:", permanent=True),
            call("progress", "- check-name: Some text (check-url)", permanent=True),
        ]
    )


def test_show_linters_lint_errors_forced(basic_project, emitter, config):
    """Show the linting results, have errors but the packing is forced."""
    builder = get_builder(config, force=True)

    # fake result from the analyzer
    linting_results = [
        linters.CheckResult(
            name="check-name",
            check_type=linters.CheckType.LINT,
            url="check-url",
            text="Some text",
            result=LintResult.ERRORS,
        ),
    ]

    builder.show_linting_results(linting_results)

    emitter.assert_interactions(
        [
            call("progress", "Lint Errors:", permanent=True),
            call("progress", "- check-name: Some text (check-url)", permanent=True),
            call("progress", "Packing anyway as requested.", permanent=True),
        ]
    )


@pytest.mark.parametrize("force", [True, False])
@pytest.mark.parametrize("destructive_mode", [True, False])
@pytest.mark.parametrize("base_indeces", [[], [1], [1, 2, 3, 4, 5]])
def test_get_charm_pack_args(config, force, base_indeces, destructive_mode):
    builder = get_builder(config, force=force)

    actual = builder._get_charm_pack_args(base_indeces, destructive_mode)

    assert actual[:3] == ["charmcraft", "pack", "--verbose"]
    assert ("--force" in actual) == force
    assert ("--destructive-mode" in actual) == destructive_mode
    for index in base_indeces:
        assert f"--bases-index={index}" in actual


# --- tests for relativise helper


def test_relativise_sameparent():
    """Two files in the same dir."""
    src = pathlib.Path("/tmp/foo/bar/src.txt")
    dst = pathlib.Path("/tmp/foo/bar/dst.txt")
    rel = relativise(src, dst)
    assert rel == pathlib.Path("dst.txt")


def test_relativise_src_under():
    """The src is in subdirectory of dst's parent."""
    src = pathlib.Path("/tmp/foo/bar/baz/src.txt")
    dst = pathlib.Path("/tmp/foo/dst.txt")
    rel = relativise(src, dst)
    assert rel == pathlib.Path("../../dst.txt")


def test_relativise_dst_under():
    """The dst is in subdirectory of src's parent."""
    src = pathlib.Path("/tmp/foo/src.txt")
    dst = pathlib.Path("/tmp/foo/bar/baz/dst.txt")
    rel = relativise(src, dst)
    assert rel == pathlib.Path("bar/baz/dst.txt")


def test_relativise_different_parents_shallow():
    """Different parents for src and dst, but shallow."""
    src = pathlib.Path("/tmp/foo/bar/src.txt")
    dst = pathlib.Path("/tmp/foo/baz/dst.txt")
    rel = relativise(src, dst)
    assert rel == pathlib.Path("../baz/dst.txt")


def test_relativise_different_parents_deep():
    """Different parents for src and dst, in a deep structure."""
    src = pathlib.Path("/tmp/foo/bar1/bar2/src.txt")
    dst = pathlib.Path("/tmp/foo/baz1/baz2/baz3/dst.txt")
    rel = relativise(src, dst)
    assert rel == pathlib.Path("../../baz1/baz2/baz3/dst.txt")


def test_format_charm_file_name_basic():
    """Basic entry."""
    bases_config = BasesConfiguration(
        **{
            "build-on": [],
            "run-on": [Base(name="xname", channel="xchannel", architectures=["xarch1"])],
        }
    )

    assert (
        format_charm_file_name("charm-name", bases_config)
        == "charm-name_xname-xchannel-xarch1.charm"
    )


def test_format_charm_file_name_multi_arch():
    """Multiple architectures."""
    bases_config = BasesConfiguration(
        **{
            "build-on": [],
            "run-on": [
                Base(
                    name="xname",
                    channel="xchannel",
                    architectures=["xarch1", "xarch2", "xarch3"],
                )
            ],
        }
    )

    assert (
        format_charm_file_name("charm-name", bases_config)
        == "charm-name_xname-xchannel-xarch1-xarch2-xarch3.charm"
    )


def test_format_charm_file_name_multi_run_on():
    """Multiple run-on entries."""
    bases_config = BasesConfiguration(
        **{
            "build-on": [],
            "run-on": [
                Base(name="x1name", channel="x1channel", architectures=["x1arch"]),
                Base(
                    name="x2name",
                    channel="x2channel",
                    architectures=["x2arch1", "x2arch2"],
                ),
            ],
        }
    )

    assert (
        format_charm_file_name("charm-name", bases_config)
        == "charm-name_x1name-x1channel-x1arch_x2name-x2channel-x2arch1-x2arch2.charm"
    )


def test_launch_shell(emitter):
    """Check bash is called while Emitter is paused."""

    def fake_run(command, check, cwd):
        """MITM to verify parameters and that emitter is paused when it's called."""
        assert command == ["bash"]
        assert check is False
        assert cwd is None
        assert emitter.paused

    with mock.patch("subprocess.run", fake_run):
        launch_shell()


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
@pytest.mark.parametrize(
    ("charms", "command_args", "charm_files", "expected_calls", "expected"),
    [
        pytest.param({}, [], [], [], {}, id="empty"),
        pytest.param(
            {"test": pathlib.Path("charms/test")},
            ["pack_cmd"],
            [],
            [
                mock.call(
                    ["pack_cmd", "--project-dir=charms/test"], stdout=mock.ANY, stderr=mock.ANY
                )
            ],
            {},
            id="no_outputs",
        ),
        pytest.param(
            {"test": pathlib.Path("charms/test")},
            ["pack_cmd"],
            ["test_amd64.charm"],
            [
                mock.call(
                    ["pack_cmd", "--project-dir=charms/test"], stdout=mock.ANY, stderr=mock.ANY
                )
            ],
            {"test": pathlib.Path("test_amd64.charm").resolve()},
            id="one_correct_charm",
        ),
        pytest.param(
            {"test": pathlib.Path("charms/test")},
            ["pack_cmd"],
            ["test_amd64.charm", "where-did-this-come-from_riscv.charm"],
            [
                mock.call(
                    ["pack_cmd", "--project-dir=charms/test"], stdout=mock.ANY, stderr=mock.ANY
                )
            ],
            {"test": pathlib.Path("test_amd64.charm").resolve()},
            id="one_correct_charm",
        ),
    ],
)
def test_subprocess_pack_charms_success(
    mocker, check, charms, charm_files, command_args, expected_calls, expected
):
    mock_check_call = mocker.patch("subprocess.check_call")
    mock_check_call.side_effect = lambda *_, **__: [pathlib.Path(f).touch() for f in charm_files]

    actual = _subprocess_pack_charms(charms, command_args)

    check.equal(actual, expected)
    check.equal(mock_check_call.mock_calls, expected_calls)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Charmcraft basic pydantic model."""

import pydantic


class ModelConfigDefaults(
    pydantic.BaseModel, extra=pydantic.Extra.forbid, frozen=True, validate_all=True
):
    """Define Charmcraft's defaults for the BaseModel configuration."""


class CustomStrictStr(pydantic.StrictStr):
    """Generic class to create custom strict strings validated by pydantic."""

    @classmethod
    def __get_validators__(cls):
        """Yield the relevant validators."""
        yield from super().__get_validators__()
        yield cls.custom_validate


class RelativePath(CustomStrictStr):
    """Constrained string which must be a relative path."""

    @classmethod
    def custom_validate(cls, value: str) -> str:
        """Validate relative path.

        Check if it's an absolute path using POSIX's '/' (not os.path.sep, as the charm's
        config is independent of the platform where charmcraft is running.
        """
        if not value:
            raise ValueError(f"{value!r} must be a valid relative path (cannot be empty)")

        if value[0] == "/":
            raise ValueError(f"{value!r} must be a valid relative path (cannot start with '/')")

        return value


class AttributeName(CustomStrictStr):
    """Constrained string that must match the name of an attribute from linters.CHECKERS."""

    @classmethod
    def custom_validate(cls, value: str) -> str:
        """Validate attribute name."""
        from charmcraft import linters  # import here to avoid cyclic imports

        valid_names = [
            checker.name
            for checker in linters.CHECKERS
            if checker.check_type == linters.CheckType.ATTRIBUTE
        ]
        if value not in valid_names:
            raise ValueError(f"Bad attribute name {value!r}")
        return value


class LinterName(CustomStrictStr):
    """Constrained string that must match the name of a linter from linters.CHECKERS."""

    @classmethod
    def custom_validate(cls, value: str) -> str:
        """Validate attribute name."""
        from charmcraft import linters  # import here to avoid cyclic imports

        valid_names = [
            checker.name
            for checker in linters.CHECKERS
            if checker.check_type == linters.CheckType.LINT
        ]
        if value not in valid_names:
            raise ValueError(f"Bad lint name {value!r}")
        return value



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
import os
from textwrap import dedent

import yaml

from charmcraft.config import load
from charmcraft.const import (
    JUJU_ACTIONS_FILENAME,
    JUJU_CONFIG_FILENAME,
    METADATA_FILENAME,
)
from charmcraft.metafiles.actions import create_actions_yaml
from charmcraft.metafiles.config import create_config_yaml
from charmcraft.metafiles.metadata import create_metadata_yaml


def test_dump_metadata_from_charmcraft_yaml(tmp_path, prepare_charmcraft_yaml):
    """Dump a metadata.yaml with full metadata. (Spec ST087)"""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description

            bases:
              - name: test-name
                channel: test-channel

            assumes:
              - test-feature
              - any-of:
                  - extra-feature-1
                  - extra-feature-2
                  - all-of:
                    - juju >= 2.9.44
                    - juju < 3
                  - all-of:
                    - juju >= 3.1.6
                    - juju < 4
              - all-of:
                  - test-feature-1
                  - test-feature-2

            containers:
              container-1:
                resource: resource-1
                bases:
                  - name: ubuntu
                    channel: 22.04
                    architectures:
                      - x86_64
                mounts:
                  - storage: storage-1
                    location: /var/lib/storage-1
              container-2:
                resource: resource-2
                bases:
                  - name: ubuntu
                    channel: 22.04
                    architectures:
                      - x86_64
                mounts:
                  - storage: storage-2
                    location: /var/lib/storage-2

            devices:
              test-device-1:
                  type: gpu
                  description: gpu
                  countmin: 1
                  countmax: 10

            title: test-title

            extra-bindings:
              test-binding-1: binding-1

            links:
              issues: https://example.com/issues
              contact:
                - https://example.com/contact
                - contact@example.com
                - "IRC #example"
              documentation: https://example.com/docs
              source:
                - https://example.com/source
                - https://example.com/source2
                - https://example.com/source3
              website:
                - https://example.com/

            peers:
              peer-1:
                interface: eth0
                limit: 1
                optional: true
                scope: global

            provides:
              provide-1:
                interface: eht1
                limit: 1
                optional: true
                scope: global

            requires:
              peer-1:
                interface: eth0
                limit: 1
                optional: true
                scope: global

            resources:
                resource-1:
                type: file
                description: resource-1
                filename: /path/to/resource-1

            storage:
              storage-1:
                type: filesystem
                description: storage-1
                location: /var/lib/storage-1
                shared: true
                read-only: false
                multiple: 5G
                minimum-size: 5G
                properties:
                  - transient

            subordinate: true

            terms:
              - https://example.com/terms
              - https://example.com/terms2
            """
        ),
    )

    config = load(tmp_path)

    create_metadata_yaml(tmp_path, config)

    metadata = yaml.safe_load((tmp_path / METADATA_FILENAME).read_text())

    assert metadata == {
        "name": "test-charm-name",
        "summary": "test-summary",
        "description": "test-description",
        "assumes": [
            "test-feature",
            {
                "any-of": [
                    "extra-feature-1",
                    "extra-feature-2",
                    {"all-of": ["juju >= 2.9.44", "juju < 3"]},
                    {"all-of": ["juju >= 3.1.6", "juju < 4"]},
                ]
            },
            {"all-of": ["test-feature-1", "test-feature-2"]},
        ],
        "containers": {
            "container-1": {
                "resource": "resource-1",
                "bases": [{"name": "ubuntu", "channel": 22.04, "architectures": ["x86_64"]}],
                "mounts": [{"storage": "storage-1", "location": "/var/lib/storage-1"}],
            },
            "container-2": {
                "resource": "resource-2",
                "bases": [{"name": "ubuntu", "channel": 22.04, "architectures": ["x86_64"]}],
                "mounts": [{"storage": "storage-2", "location": "/var/lib/storage-2"}],
            },
        },
        "devices": {
            "test-device-1": {"type": "gpu", "description": "gpu", "countmin": 1, "countmax": 10}
        },
        "display-name": "test-title",
        "peers": {
            "peer-1": {"interface": "eth0", "limit": 1, "optional": True, "scope": "global"}
        },
        "provides": {
            "provide-1": {"interface": "eht1", "limit": 1, "optional": True, "scope": "global"}
        },
        "requires": {
            "peer-1": {"interface": "eth0", "limit": 1, "optional": True, "scope": "global"}
        },
        "resources": {
            "resource-1": None,
            "type": "file",
            "description": "resource-1",
            "filename": "/path/to/resource-1",
        },
        "storage": {
            "storage-1": {
                "type": "filesystem",
                "description": "storage-1",
                "location": "/var/lib/storage-1",
                "shared": True,
                "read-only": False,
                "multiple": "5G",
                "minimum-size": "5G",
                "properties": ["transient"],
            }
        },
        "subordinate": True,
        "terms": ["https://example.com/terms", "https://example.com/terms2"],
        "extra-bindings": {"test-binding-1": "binding-1"},
        "docs": "https://example.com/docs",
        "issues": "https://example.com/issues",
        "maintainers": ["https://example.com/contact", "contact@example.com", "IRC #example"],
        "source": [
            "https://example.com/source",
            "https://example.com/source2",
            "https://example.com/source3",
        ],
        "website": ["https://example.com/"],
    }


def test_copy_metadata_from_metadata_yaml(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml
):
    """Copy a metadata.yaml with full metadata. (Spec ST087)"""
    prepare_charmcraft_yaml(
        dedent(
            """
            type: charm
            bases:
              - name: test-name
                channel: test-channel
            """
        )
    )
    prepare_metadata_yaml(
        dedent(
            """
            name: test-charm-name
            summary: test-summary
            description: test-description
            assumes:
              - test-feature
              - any-of:
                  - extra-feature-1
                  - extra-feature-2
                  - all-of:
                    - juju >= 2.9.44
                    - juju < 3
                  - all-of:
                    - juju >= 3.1.6
                    - juju < 4
              - all-of:
                  - test-feature-1
                  - test-feature-2

            containers:
              container-1:
                resource: resource-1
                bases:
                  - name: ubuntu
                    channel: 22.04
                    architectures:
                      - x86_64
                mounts:
                  - storage: storage-1
                    location: /var/lib/storage-1
              container-2:
                resource: resource-2
                bases:
                  - name: ubuntu
                    channel: 22.04
                    architectures:
                      - x86_64
                mounts:
                  - storage: storage-2
                    location: /var/lib/storage-2

            devices:
              test-device-1:
                  type: gpu
                  description: gpu
                  countmin: 1
                  countmax: 10

            display-name: test-title

            docs: https://example.com/docs

            extra-bindings:
              test-binding-1: binding-1

            issues: https://example.com/issues

            maintainers:
              - https://example.com/contact
              - contact@example.com
              - "IRC #example"

            peers:
              peer-1:
                interface: eth0
                limit: 1
                optional: true
                scope: global

            provides:
              provide-1:
                interface: eht1
                limit: 1
                optional: true
                scope: global

            requires:
              peer-1:
                interface: eth0
                limit: 1
                optional: true
                scope: global

            resources:
                resource-1:
                type: file
                description: resource-1
                filename: /path/to/resource-1

            source:
              - https://example.com/source
              - https://example.com/source2
              - https://example.com/source3

            storage:
              storage-1:
                type: filesystem
                description: storage-1
                location: /var/lib/storage-1
                shared: true
                read-only: false
                multiple: 5G
                minimum-size: 5G
                properties:
                  - transient

            subordinate: true

            terms:
              - https://example.com/terms
              - https://example.com/terms2

            website:
              - https://example.com/

            #### TEST-COPY ####
            """
        ),
    )

    config = load(tmp_path)

    os.mkdir(tmp_path / "new")

    create_metadata_yaml(tmp_path / "new", config)

    metadata_yaml = (tmp_path / "new" / METADATA_FILENAME).read_text()

    # Copy will preserve the TEST-COPY comment
    assert "TEST-COPY" in metadata_yaml

    metadata = yaml.safe_load(metadata_yaml)

    assert metadata == {
        "name": "test-charm-name",
        "summary": "test-summary",
        "description": "test-description",
        "assumes": [
            "test-feature",
            {
                "any-of": [
                    "extra-feature-1",
                    "extra-feature-2",
                    {"all-of": ["juju >= 2.9.44", "juju < 3"]},
                    {"all-of": ["juju >= 3.1.6", "juju < 4"]},
                ]
            },
            {"all-of": ["test-feature-1", "test-feature-2"]},
        ],
        "containers": {
            "container-1": {
                "resource": "resource-1",
                "bases": [{"name": "ubuntu", "channel": 22.04, "architectures": ["x86_64"]}],
                "mounts": [{"storage": "storage-1", "location": "/var/lib/storage-1"}],
            },
            "container-2": {
                "resource": "resource-2",
                "bases": [{"name": "ubuntu", "channel": 22.04, "architectures": ["x86_64"]}],
                "mounts": [{"storage": "storage-2", "location": "/var/lib/storage-2"}],
            },
        },
        "devices": {
            "test-device-1": {"type": "gpu", "description": "gpu", "countmin": 1, "countmax": 10}
        },
        "display-name": "test-title",
        "peers": {
            "peer-1": {"interface": "eth0", "limit": 1, "optional": True, "scope": "global"}
        },
        "provides": {
            "provide-1": {"interface": "eht1", "limit": 1, "optional": True, "scope": "global"}
        },
        "requires": {
            "peer-1": {"interface": "eth0", "limit": 1, "optional": True, "scope": "global"}
        },
        "resources": {
            "resource-1": None,
            "type": "file",
            "description": "resource-1",
            "filename": "/path/to/resource-1",
        },
        "storage": {
            "storage-1": {
                "type": "filesystem",
                "description": "storage-1",
                "location": "/var/lib/storage-1",
                "shared": True,
                "read-only": False,
                "multiple": "5G",
                "minimum-size": "5G",
                "properties": ["transient"],
            }
        },
        "subordinate": True,
        "terms": ["https://example.com/terms", "https://example.com/terms2"],
        "extra-bindings": {"test-binding-1": "binding-1"},
        "docs": "https://example.com/docs",
        "issues": "https://example.com/issues",
        "maintainers": ["https://example.com/contact", "contact@example.com", "IRC #example"],
        "source": [
            "https://example.com/source",
            "https://example.com/source2",
            "https://example.com/source3",
        ],
        "website": ["https://example.com/"],
    }


def test_copy_metadata_from_metadata_yaml_with_arbitrary_keys(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml
):
    """Copy a metadata.yaml with full metadata. (Spec ST087)"""
    prepare_charmcraft_yaml(
        dedent(
            """
            type: charm
            bases:
              - name: test-name
                channel: test-channel
            """
        )
    )
    prepare_metadata_yaml(
        dedent(
            """
            name: test-charm-name
            summary: test-summary
            description: test-description
            assumes:
              - test-feature
              - any-of:
                  - extra-feature-1
                  - extra-feature-2
              - all-of:
                  - test-feature-1
                  - test-feature-2

            containers:
              container-1:
                resource: resource-1
                bases:
                  - name: ubuntu
                    channel: 22.04
                    architectures:
                      - x86_64
                mounts:
                  - storage: storage-1
                    location: /var/lib/storage-1
              container-2:
                resource: resource-2
                bases:
                  - name: ubuntu
                    channel: 22.04
                    architectures:
                      - x86_64
                mounts:
                  - storage: storage-2
                    location: /var/lib/storage-2

            devices:
              test-device-1:
                  type: gpu
                  description: gpu
                  countmin: 1
                  countmax: 10

            display-name: test-title

            docs: https://example.com/docs

            extra-bindings:
              test-binding-1: binding-1

            issues: https://example.com/issues

            maintainers:
              - https://example.com/contact
              - contact@example.com
              - "IRC #example"

            peers:
              peer-1:
                interface: eth0
                limit: 1
                optional: true
                scope: global

            provides:
              provide-1:
                interface: eht1
                limit: 1
                optional: true
                scope: global

            requires:
              peer-1:
                interface: eth0
                limit: 1
                optional: true
                scope: global

            resources:
                resource-1:
                type: file
                description: resource-1
                filename: /path/to/resource-1

            source:
              - https://example.com/source
              - https://example.com/source2
              - https://example.com/source3

            storage:
              storage-1:
                type: filesystem
                description: storage-1
                location: /var/lib/storage-1
                shared: true
                read-only: false
                multiple: 5G
                minimum-size: 5G
                properties:
                  - transient

            subordinate: true

            terms:
              - https://example.com/terms
              - https://example.com/terms2

            website:
              - https://example.com/

            test-arbitrary-key: test-arbitrary-value
            test-arbitrary-key-2: test-arbitrary-value-2

            #### TEST-COPY ####
            """
        ),
    )

    config = load(tmp_path)

    os.mkdir(tmp_path / "new")

    create_metadata_yaml(tmp_path / "new", config)

    metadata_yaml = (tmp_path / "new" / METADATA_FILENAME).read_text()

    # Copy will preserve the TEST-COPY comment
    assert "TEST-COPY" in metadata_yaml

    metadata = yaml.safe_load(metadata_yaml)

    assert metadata == {
        "name": "test-charm-name",
        "summary": "test-summary",
        "description": "test-description",
        "assumes": [
            "test-feature",
            {"any-of": ["extra-feature-1", "extra-feature-2"]},
            {"all-of": ["test-feature-1", "test-feature-2"]},
        ],
        "containers": {
            "container-1": {
                "resource": "resource-1",
                "bases": [{"name": "ubuntu", "channel": 22.04, "architectures": ["x86_64"]}],
                "mounts": [{"storage": "storage-1", "location": "/var/lib/storage-1"}],
            },
            "container-2": {
                "resource": "resource-2",
                "bases": [{"name": "ubuntu", "channel": 22.04, "architectures": ["x86_64"]}],
                "mounts": [{"storage": "storage-2", "location": "/var/lib/storage-2"}],
            },
        },
        "devices": {
            "test-device-1": {"type": "gpu", "description": "gpu", "countmin": 1, "countmax": 10}
        },
        "display-name": "test-title",
        "peers": {
            "peer-1": {"interface": "eth0", "limit": 1, "optional": True, "scope": "global"}
        },
        "provides": {
            "provide-1": {"interface": "eht1", "limit": 1, "optional": True, "scope": "global"}
        },
        "requires": {
            "peer-1": {"interface": "eth0", "limit": 1, "optional": True, "scope": "global"}
        },
        "resources": {
            "resource-1": None,
            "type": "file",
            "description": "resource-1",
            "filename": "/path/to/resource-1",
        },
        "storage": {
            "storage-1": {
                "type": "filesystem",
                "description": "storage-1",
                "location": "/var/lib/storage-1",
                "shared": True,
                "read-only": False,
                "multiple": "5G",
                "minimum-size": "5G",
                "properties": ["transient"],
            }
        },
        "subordinate": True,
        "terms": ["https://example.com/terms", "https://example.com/terms2"],
        "extra-bindings": {"test-binding-1": "binding-1"},
        "docs": "https://example.com/docs",
        "issues": "https://example.com/issues",
        "maintainers": ["https://example.com/contact", "contact@example.com", "IRC #example"],
        "source": [
            "https://example.com/source",
            "https://example.com/source2",
            "https://example.com/source3",
        ],
        "website": ["https://example.com/"],
        "test-arbitrary-key": "test-arbitrary-value",
        "test-arbitrary-key-2": "test-arbitrary-value-2",
    }


def test_copy_bundle_metadata_from_metadata_yaml(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml
):
    """Copy a metadata.yaml when type is bundle."""
    prepare_charmcraft_yaml(
        dedent(
            """
            type: bundle
            """
        ),
    )
    prepare_metadata_yaml(
        dedent(
            """
            name: test-charm-name
            description: This is a test bundle.

            variables:
              data-port:           &data-port            br-ex:eno2
              worker-multiplier:   &worker-multiplier    0.25


            series: bionic

            tags: [monitoring]

            tags: [database, utility]


            applications:
              easyrsa:
                charm: containers-easyrsa
                revision: 8
                channel: latest/edge
                series: bionic

                resources:
                  easyrsa: 5

                resources:
                  easyrsa: ./relative/path/to/file

                resources:
                  easyrsa: /absolute/path/to/file

                num_units: 2

                to: 3, new
                to: ["django/0", "django/1", "django/2"]
                to: ["django"]
                to: ["lxd"]
                to: ["lxd:2", "lxd:3"]
                to: ["lxd:nova-compute/2", "lxd:glance/3"]

                expose: true

                offers:
                  my-offer:
                    endpoints:
                    - apache-website
                    acl:
                      admin: admin
                      user1: read

                options:
                  osd-devices: /dev/sdb
                  worker-multiplier: *worker-multiplier

                annotations:
                  gui-x: 450
                  gui-y: 550

                constraints: root-disk=8G

                constraints: cores=4 mem=4G root-disk=16G

                constraints: zones=us-east-1a

                storage:
                  database: ebs,10G,1

                bindings:
                  "": alpha
                  kube-api-endpoint: internal
                  loadbalancer: dmz

                plan: acme-support/default

            machines:
              "1":
              "2":
                series: bionic
                constraints: cores=2 mem=2G
              "3":
                constraints: cores=3 root-disk=1T

            relations:
            - - kubernetes-master:kube-api-endpoint
              - kubeapi-load-balancer:apiserver
            - - kubernetes-master:loadbalancer
              - kubeapi-load-balancer:loadbalancer

            saas:
              svc1:
                url: localoffer.svc1
              svc2:
                url: admin/localoffer.svc2
              svc3:
                url: othercontroller:admin/offer.svc3

            #### TEST-COPY ####
            """
        ),
    )

    config = load(tmp_path)

    os.mkdir(tmp_path / "new")

    create_metadata_yaml(tmp_path / "new", config)

    metadata_yaml = (tmp_path / "new" / METADATA_FILENAME).read_text()

    # Copy will preserve the TEST-COPY comment
    assert "TEST-COPY" in metadata_yaml


def test_dump_actions_from_charmcraft_yaml(tmp_path, prepare_charmcraft_yaml):
    """Dump a actions.yaml from charmcraft.yaml."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description

            bases:
              - name: test-name
                channel: test-channel

            actions:
              pause:
                description: Pause the database.
              resume:
                description: Resume a paused database.
              snapshot:
                description: Take a snapshot of the database.
                params:
                  filename:
                    type: string
                    description: The name of the snapshot file.
                  compression:
                    type: object
                    description: The type of compression to use.
                    properties:
                      kind:
                        type: string
                        enum: [gzip, bzip2, xz]
                      quality:
                        description: Compression quality
                        type: integer
                        minimum: 0
                        maximum: 9
                required: [filename]
                additionalProperties: false
            """
        ),
    )

    config = load(tmp_path)

    create_actions_yaml(tmp_path, config)

    actions = yaml.safe_load((tmp_path / JUJU_ACTIONS_FILENAME).read_text())

    assert actions == {
        "pause": {"description": "Pause the database."},
        "resume": {"description": "Resume a paused database."},
        "snapshot": {
            "description": "Take a snapshot of the database.",
            "params": {
                "filename": {
                    "type": "string",
                    "description": "The name of the snapshot file.",
                },
                "compression": {
                    "type": "object",
                    "description": "The type of compression to use.",
                    "properties": {
                        "kind": {"type": "string", "enum": ["gzip", "bzip2", "xz"]},
                        "quality": {
                            "description": "Compression quality",
                            "type": "integer",
                            "minimum": 0,
                            "maximum": 9,
                        },
                    },
                },
            },
            "required": ["filename"],
            "additionalProperties": False,
        },
    }


def test_copy_actions_from_actions_yaml(tmp_path, prepare_charmcraft_yaml, prepare_actions_yaml):
    """Dump a actions.yaml from charmcraft.yaml."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description
            """
        ),
    )
    prepare_actions_yaml(
        dedent(
            """
            pause:
              description: Pause the database.
            resume:
              description: Resume a paused database.
            snapshot:
              description: Take a snapshot of the database.
              params:
                filename:
                  type: string
                  description: The name of the snapshot file.
                compression:
                  type: object
                  description: The type of compression to use.
                  properties:
                    kind:
                      type: string
                      enum: [gzip, bzip2, xz]
                    quality:
                      description: Compression quality
                      type: integer
                      minimum: 0
                      maximum: 9
              required: [filename]
              additionalProperties: false
            #### TEST-COPY ####
            """
        ),
    )

    config = load(tmp_path)

    os.mkdir(tmp_path / "new")

    create_actions_yaml(tmp_path / "new", config)

    actions_yaml = (tmp_path / "new" / JUJU_ACTIONS_FILENAME).read_text()

    # Copy will preserve the TEST-COPY comment
    assert "TEST-COPY" in actions_yaml

    actions = yaml.safe_load(actions_yaml)

    assert actions == {
        "pause": {"description": "Pause the database."},
        "resume": {"description": "Resume a paused database."},
        "snapshot": {
            "description": "Take a snapshot of the database.",
            "params": {
                "filename": {
                    "type": "string",
                    "description": "The name of the snapshot file.",
                },
                "compression": {
                    "type": "object",
                    "description": "The type of compression to use.",
                    "properties": {
                        "kind": {"type": "string", "enum": ["gzip", "bzip2", "xz"]},
                        "quality": {
                            "description": "Compression quality",
                            "type": "integer",
                            "minimum": 0,
                            "maximum": 9,
                        },
                    },
                },
            },
            "required": ["filename"],
            "additionalProperties": False,
        },
    }


def test_dump_config_from_charmcraft_yaml(tmp_path, prepare_charmcraft_yaml):
    """Dump a config.yaml from charmcraft.yaml."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description

            config:
              options:
                test-int:
                  default: 123
                  description: test-1
                  type: int
                test-string:
                  description: test-2
                  type: string
                test-float:
                  default: 1.23
                  type: float
                test-bool:
                  default: true
                  type: boolean
            """
        ),
    )

    config = load(tmp_path)

    create_config_yaml(tmp_path, config)

    config_data = yaml.safe_load((tmp_path / JUJU_CONFIG_FILENAME).read_text())

    assert config_data == {
        "options": {
            "test-int": {"default": 123, "description": "test-1", "type": "int"},
            "test-string": {"description": "test-2", "type": "string"},
            "test-float": {"default": 1.23, "type": "float"},
            "test-bool": {"default": True, "type": "boolean"},
        },
    }


def test_copy_config_from_config_yaml(tmp_path, prepare_charmcraft_yaml, prepare_config_yaml):
    """Dump a actions.yaml from charmcraft.yaml."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description
            """
        ),
    )
    prepare_config_yaml(
        dedent(
            """
            options:
              test-int:
                default: 123
                description: test-1
                type: int
              test-string:
                description: test-2
                type: string
              test-float:
                default: 1.23
                type: float
              test-bool:
                default: true
                type: boolean
            #### TEST-COPY ####
            """
        ),
    )

    config = load(tmp_path)

    os.mkdir(tmp_path / "new")

    create_config_yaml(tmp_path / "new", config)

    config_yaml = (tmp_path / "new" / JUJU_CONFIG_FILENAME).read_text()

    # Copy will preserve the TEST-COPY comment
    assert "TEST-COPY" in config_yaml

    config_data = yaml.safe_load(config_yaml)

    assert config_data == {
        "options": {
            "test-int": {"default": 123, "description": "test-1", "type": "int"},
            "test-string": {"description": "test-2", "type": "string"},
            "test-float": {"default": 1.23, "type": "float"},
            "test-bool": {"default": True, "type": "boolean"},
        },
    }



# Copyright 2021-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Tests for analyze and lint code."""

import pathlib
import sys
from textwrap import dedent
from unittest.mock import patch

import pytest

from charmcraft.linters import (
    CHECKERS,
    BaseChecker,
    CheckType,
    Entrypoint,
    Framework,
    JujuActions,
    JujuConfig,
    JujuMetadata,
    Language,
    analyze,
    check_dispatch_with_python_entrypoint,
    get_entrypoint_from_dispatch,
)
from charmcraft.models.lint import LintResult

EXAMPLE_DISPATCH = """
#!/bin/sh

PYTHONPATH=lib:venv ./charm.py
"""

# --- tests for helper functions


def test_epfromdispatch_ok(tmp_path):
    """An entrypoint is found in the dispatch."""
    dispatch = tmp_path / "dispatch"
    dispatch.write_text(EXAMPLE_DISPATCH)
    entrypoint = tmp_path / "charm.py"
    result = get_entrypoint_from_dispatch(tmp_path)
    assert result == entrypoint


def test_epfromdispatch_no_dispatch(tmp_path):
    """The charm has no dispatch at all."""
    result = get_entrypoint_from_dispatch(tmp_path)
    assert result is None


def test_epfromdispatch_inaccessible_dispatch(tmp_path):
    """The charm has a dispatch we can't use."""
    dispatch = tmp_path / "dispatch"
    dispatch.touch()
    dispatch.chmod(0o000)
    result = get_entrypoint_from_dispatch(tmp_path)
    assert result is None


def test_epfromdispatch_broken_dispatch(tmp_path):
    """The charm has a dispatch which we can't decode."""
    dispatch = tmp_path / "dispatch"
    dispatch.write_bytes(b"\xC0\xC0")
    result = get_entrypoint_from_dispatch(tmp_path)
    assert result is None


def test_epfromdispatch_empty_dispatch(tmp_path):
    """The charm dispatch is empty."""
    dispatch = tmp_path / "dispatch"
    dispatch.write_text("")
    result = get_entrypoint_from_dispatch(tmp_path)
    assert result is None


def test_checkdispatchpython_python_ok(tmp_path):
    """The charm is written in Python."""
    entrypoint = tmp_path / "charm.py"
    entrypoint.touch(mode=0o700)
    with patch("charmcraft.linters.get_entrypoint_from_dispatch", return_value=entrypoint):
        result = check_dispatch_with_python_entrypoint(tmp_path)
    assert result == entrypoint


def test_checkdispatchpython_no_entrypoint(tmp_path):
    """Cannot find the entrypoint used in dispatch."""
    entrypoint = tmp_path / "charm.py"
    with patch("charmcraft.linters.get_entrypoint_from_dispatch", return_value=entrypoint):
        result = check_dispatch_with_python_entrypoint(tmp_path)
    assert result is None


def test_checkdispatchpython_nothing_from_dispatch(tmp_path):
    """The dispatch is not there, or not usable."""
    with patch("charmcraft.linters.get_entrypoint_from_dispatch", return_value=None):
        result = check_dispatch_with_python_entrypoint(tmp_path)
    assert result is None


def test_checkdispatchpython_entrypoint_is_not_python(tmp_path):
    """The charm entrypoint has not a .py extension."""
    dispatch = tmp_path / "dispatch"
    dispatch.write_text(
        """
        #!/bin/sh
        ./charm
    """
    )
    entrypoint = tmp_path / "charm"
    entrypoint.touch(mode=0o700)
    with patch("charmcraft.linters.get_entrypoint_from_dispatch", return_value=entrypoint):
        result = check_dispatch_with_python_entrypoint(tmp_path)
    assert result is None


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_checkdispatchpython_entrypoint_no_exec(tmp_path):
    """The charm entrypoint is not executable."""
    dispatch = tmp_path / "dispatch"
    dispatch.write_text(EXAMPLE_DISPATCH)
    entrypoint = tmp_path / "charm.py"
    entrypoint.touch()
    with patch("charmcraft.linters.get_entrypoint_from_dispatch", return_value=entrypoint):
        result = check_dispatch_with_python_entrypoint(tmp_path)
    assert result is None


# --- tests for Language checker


def test_language_python():
    """The charm is written in Python."""
    with patch("charmcraft.linters.check_dispatch_with_python_entrypoint") as mock_check:
        mock_check.return_value = pathlib.Path("entrypoint")
        result = Language().run(pathlib.Path("somedir"))
    assert result == Language.Result.PYTHON
    mock_check.assert_called_with(pathlib.Path("somedir"))


def test_language_no_dispatch(tmp_path):
    """The charm has no dispatch at all."""
    with patch("charmcraft.linters.check_dispatch_with_python_entrypoint") as mock_check:
        mock_check.return_value = None
        result = Language().run(pathlib.Path("somedir"))
    assert result == Language.Result.UNKNOWN
    mock_check.assert_called_with(pathlib.Path("somedir"))


# --- tests for Framework checker


def test_framework_run_operator():
    """Check for Operator Framework was successful."""
    checker = Framework()
    with patch.object(Framework, "_check_operator", lambda self, path: True):
        result = checker.run(pathlib.Path("somepath"))
    assert result == Framework.Result.OPERATOR
    assert checker.text == "The charm is based on the Operator Framework."


def test_framework_run_reactive():
    """Check for Reactive Framework was successful."""
    checker = Framework()
    with patch.object(Framework, "_check_operator", lambda self, path: False):
        with patch.object(Framework, "_check_reactive", lambda self, path: True):
            result = checker.run(pathlib.Path("somepath"))
    assert result == Framework.Result.REACTIVE
    assert checker.text == "The charm is based on the Reactive Framework."


def test_framework_run_unknown():
    """No check for any framework was successful."""
    checker = Framework()
    with patch.object(Framework, "_check_operator", lambda self, path: False):
        with patch.object(Framework, "_check_reactive", lambda self, path: False):
            result = checker.run(pathlib.Path("somepath"))
    assert result == Framework.Result.UNKNOWN
    assert checker.text == "The charm is not based on any known Framework."


@pytest.mark.parametrize(
    "import_line",
    [
        "import ops",
        "import stuff, ops, morestuff",
        "from ops import charm",
        "from ops.charm import CharmBase",
    ],
)
def test_framework_operator_used_ok(tmp_path, import_line):
    """All conditions for 'framework' are in place."""
    # an entry point that import ops
    entrypoint = tmp_path / "charm.py"
    entrypoint.write_text(f"{import_line}")

    # an ops directory inside venv
    opsdir = tmp_path / "venv" / "ops"
    opsdir.mkdir(parents=True)

    # check
    with patch("charmcraft.linters.check_dispatch_with_python_entrypoint") as mock_check:
        mock_check.return_value = pathlib.Path(entrypoint)
        result = Framework()._check_operator(tmp_path)
    assert result is True
    mock_check.assert_called_with(tmp_path)


def test_framework_operator_language_not_python(tmp_path):
    """The language trait is not set to Python."""
    # an entry point that is not really python
    entrypoint = tmp_path / "charm.py"
    entrypoint.write_text("no python :)")

    # an ops directory inside venv
    opsdir = tmp_path / "venv" / "ops"
    opsdir.mkdir(parents=True)

    # check
    with patch("charmcraft.linters.check_dispatch_with_python_entrypoint") as mock_check:
        mock_check.return_value = None
        result = Framework()._check_operator(tmp_path)
    assert result is False


def test_framework_operator_venv_directory_missing(tmp_path):
    """The charm has not a specific 'venv' dir."""
    # an entry point that import ops
    entrypoint = tmp_path / "charm.py"
    entrypoint.write_text("import ops")

    # check
    with patch("charmcraft.linters.check_dispatch_with_python_entrypoint") as mock_check:
        mock_check.return_value = pathlib.Path(entrypoint)
        result = Framework()._check_operator(tmp_path)
    assert result is False


def test_framework_operator_no_venv_ops_directory(tmp_path):
    """The charm *has not* a specific 'venv/ops' dir."""
    # an entry point that import ops
    entrypoint = tmp_path / "charm.py"
    entrypoint.write_text("import ops")

    # an empty venv
    venvdir = tmp_path / "venv"
    venvdir.mkdir()

    # check
    with patch("charmcraft.linters.check_dispatch_with_python_entrypoint") as mock_check:
        mock_check.return_value = pathlib.Path(entrypoint)
        result = Framework()._check_operator(tmp_path)
    assert result is False


def test_framework_operator_venv_ops_directory_is_not_a_dir(tmp_path):
    """The charm has not a specific 'venv/ops' *dir*."""
    # an entry point that import ops
    entrypoint = tmp_path / "charm.py"
    entrypoint.write_text("import ops")

    # an ops *file* inside venv
    opsfile = tmp_path / "venv" / "ops"
    opsfile.parent.mkdir()
    opsfile.touch()

    # check
    with patch("charmcraft.linters.check_dispatch_with_python_entrypoint") as mock_check:
        mock_check.return_value = pathlib.Path(entrypoint)
        result = Framework()._check_operator(tmp_path)
    assert result is False


def test_framework_operator_corrupted_entrypoint(tmp_path):
    """Cannot parse the Python file."""
    # an entry point that import ops
    entrypoint = tmp_path / "charm.py"
    entrypoint.write_text("xx --")  # not really Python

    # an ops directory inside venv
    opsdir = tmp_path / "venv" / "ops"
    opsdir.mkdir(parents=True)

    # check
    with patch("charmcraft.linters.check_dispatch_with_python_entrypoint") as mock_check:
        mock_check.return_value = pathlib.Path(entrypoint)
        result = Framework()._check_operator(tmp_path)
    assert result is False


@pytest.mark.parametrize(
    "import_line",
    [
        "import logging",
        "import whatever.ops",
        "from stuff import ops",
        "from stuff.ops import whatever",
    ],
)
def test_framework_operator_no_ops_imported(tmp_path, monkeypatch, import_line):
    """Different imports that are NOT importing the Operator Framework."""
    # an entry point that import ops
    entrypoint = tmp_path / "charm.py"
    entrypoint.write_text(f"{import_line}")

    # an ops directory inside venv
    opsdir = tmp_path / "venv" / "ops"
    opsdir.mkdir(parents=True)

    # check
    with patch("charmcraft.linters.check_dispatch_with_python_entrypoint") as mock_check:
        mock_check.return_value = pathlib.Path(entrypoint)
        result = Framework()._check_operator(tmp_path)
    assert result is False


@pytest.mark.parametrize(
    "import_line",
    [
        "import charms.reactive",
        "import stuff, charms.reactive, morestuff",
        "from charms.reactive import stuff",
        "from charms.reactive.stuff import Stuff",
    ],
)
@pytest.mark.parametrize(
    ("charm_module", "charmcraft_yaml", "metadata_yaml"),
    [
        (
            "foobar.py",
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: foobar
                summary: Small text.
                description: Lot of text.
                """
            ),
        ),
        (
            "foo_bar.py",
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: foo-bar
                summary: Small text.
                description: Lot of text.
                """
            ),
        ),
        (
            "foobar.py",
            None,
            dedent(
                """\
                name: foobar
                summary: Small text.
                description: Lot of text.
                """
            ),
        ),
        (
            "foo_bar.py",
            None,
            dedent(
                """\
                name: foo-bar
                summary: Small text.
                description: Lot of text.
                """
            ),
        ),
    ],
)
def test_framework_reactive_used_ok(
    tmp_path,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml,
    metadata_yaml,
    import_line,
    charm_module,
):
    """The reactive framework was used.

    Parametrized args:
    - import_line: different ways to express the import
    """
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    # a Python file that imports charms.reactive
    entrypoint = tmp_path / "reactive" / charm_module
    entrypoint.parent.mkdir()
    entrypoint.write_text(f"{import_line}")

    # the reactive lib is used
    reactive_lib = tmp_path / "wheelhouse" / "charms.reactive-1.0.1.zip"
    reactive_lib.parent.mkdir()
    reactive_lib.touch()

    # check
    result = Framework()._check_reactive(tmp_path)
    assert result is True


def test_framework_reactive_no_metadata(tmp_path, monkeypatch):
    """No useful name from metadata."""
    # a Python file that imports charms.reactive
    entrypoint = tmp_path / "reactive" / "foobar.py"
    entrypoint.parent.mkdir()
    entrypoint.write_text("import charms.reactive")

    # the reactive lib is used
    reactive_lib = tmp_path / "wheelhouse" / "charms.reactive-1.0.1.zip"
    reactive_lib.parent.mkdir()
    reactive_lib.touch()

    # check
    result = Framework()._check_reactive(tmp_path)
    assert result is False


def test_framework_reactive_no_entrypoint(tmp_path, monkeypatch):
    """Missing entrypoint file."""
    # metadata file with needed name field
    metadata_file = tmp_path / "metadata.yaml"
    metadata_file.write_text("name: foobar")

    # the reactive lib is used
    reactive_lib = tmp_path / "wheelhouse" / "charms.reactive-1.0.1.zip"
    reactive_lib.parent.mkdir()
    reactive_lib.touch()

    # check
    result = Framework()._check_reactive(tmp_path)
    assert result is False


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_framework_reactive_unaccesible_entrypoint(tmp_path, monkeypatch):
    """Cannot read the entrypoint file."""
    # metadata file with needed name field
    metadata_file = tmp_path / "metadata.yaml"
    metadata_file.write_text("name: foobar")

    # a Python file that imports charms.reactive
    entrypoint = tmp_path / "reactive" / "foobar.py"
    entrypoint.parent.mkdir()
    entrypoint.write_text("import charms.reactive")
    entrypoint.chmod(0o000)

    # the reactive lib is used
    reactive_lib = tmp_path / "wheelhouse" / "charms.reactive-1.0.1.zip"
    reactive_lib.parent.mkdir()
    reactive_lib.touch()

    # check
    result = Framework()._check_reactive(tmp_path)
    assert result is False


def test_framework_reactive_corrupted_entrypoint(tmp_path, monkeypatch):
    """The entrypoint is not really a Python file."""
    # metadata file with needed name field
    metadata_file = tmp_path / "metadata.yaml"
    metadata_file.write_text("name: foobar")

    # a Python file that imports charms.reactive
    entrypoint = tmp_path / "reactive" / "foobar.py"
    entrypoint.parent.mkdir()
    entrypoint.write_text("xx --")  # not really Python

    # the reactive lib is used
    reactive_lib = tmp_path / "wheelhouse" / "charms.reactive-1.0.1.zip"
    reactive_lib.parent.mkdir()
    reactive_lib.touch()

    # check
    result = Framework()._check_reactive(tmp_path)
    assert result is False


def test_framework_reactive_no_wheelhouse(tmp_path, monkeypatch):
    """The wheelhouse directory does not exist."""
    # metadata file with needed name field
    metadata_file = tmp_path / "metadata.yaml"
    metadata_file.write_text("name: foobar")

    # a Python file that imports charms.reactive
    entrypoint = tmp_path / "reactive" / "foobar.py"
    entrypoint.parent.mkdir()
    entrypoint.write_text("import charms.reactive")

    # check
    result = Framework()._check_reactive(tmp_path)
    assert result is False


def test_framework_reactive_no_reactive_lib(tmp_path, monkeypatch):
    """The wheelhouse directory has no reactive lib."""
    # metadata file with needed name field
    metadata_file = tmp_path / "metadata.yaml"
    metadata_file.write_text("name: foobar")

    # a Python file that imports charms.reactive
    entrypoint = tmp_path / "reactive" / "foobar.py"
    entrypoint.parent.mkdir()
    entrypoint.write_text("import charms.reactive")

    # the reactive lib is used
    reactive_lib = tmp_path / "wheelhouse" / "charms.noreallyreactive-1.0.1.zip"
    reactive_lib.parent.mkdir()
    reactive_lib.touch()

    # check
    result = Framework()._check_reactive(tmp_path)
    assert result is False


@pytest.mark.parametrize(
    "import_line",
    [
        "import logging",
        "import whatever.charms.reactive",
        "import charms.whatever.reactive",
        "from stuff.charms import reactive",
        "from charms.stuff import reactive",
        "from stuff.charms.reactive import whatever",
    ],
)
def test_framework_reactive_no_reactive_imported(tmp_path, monkeypatch, import_line):
    """Different imports that are NOT importing the Reactive Framework."""
    # metadata file with needed name field
    metadata_file = tmp_path / "metadata.yaml"
    metadata_file.write_text("name: foobar")

    # a Python file that imports charms.reactive
    entrypoint = tmp_path / "reactive" / "foobar.py"
    entrypoint.parent.mkdir()
    entrypoint.write_text(f"{import_line}")

    # the reactive lib is used
    reactive_lib = tmp_path / "wheelhouse" / "charms.reactive-1.0.1.zip"
    reactive_lib.parent.mkdir()
    reactive_lib.touch()

    # check
    result = Framework()._check_reactive(tmp_path)
    assert result is False


# --- tests for JujuMetadata checker


def test_jujumetadata_all_ok(tmp_path):
    """All conditions ok for JujuMetadata to result ok."""
    # metadata file with proper fields
    metadata_file = tmp_path / "metadata.yaml"
    metadata_file.write_text(
        """
        name: foobar
        summary: Small text.
        description: Lot of text.
    """
    )
    result = JujuMetadata().run(tmp_path)
    assert result == JujuMetadata.Result.OK


def test_jujumetadata_missing_file(tmp_path):
    """No metadata.yaml file at all."""
    linter = JujuMetadata()
    result = linter.run(tmp_path)
    assert result == JujuMetadata.Result.ERRORS
    assert linter.text == "Cannot read the metadata.yaml file."


def test_jujumetadata_file_corrupted(tmp_path):
    """The metadata.yaml file is not valid YAML."""
    metadata_file = tmp_path / "metadata.yaml"
    metadata_file.write_text(" - \n-")
    linter = JujuMetadata()
    result = linter.run(tmp_path)
    assert result == JujuMetadata.Result.ERRORS
    assert linter.text == "The metadata.yaml file is not a valid YAML file."


_mandatory_fields = ["summary", "description", "name"]


@pytest.mark.parametrize("to_miss", range(len(_mandatory_fields)))
def test_jujumetadata_missing_field_simple(tmp_path, to_miss):
    """A required field is missing in the metadata file."""
    included_fields = _mandatory_fields.copy()
    missing = included_fields.pop(to_miss)

    # metadata file with not all fields
    metadata_file = tmp_path / "metadata.yaml"
    content = "\n".join(f"{field}: some text" for field in included_fields)
    metadata_file.write_text(content)
    linter = JujuMetadata()
    result = linter.run(tmp_path)
    assert result == JujuMetadata.Result.ERRORS
    assert linter.text == (
        f"The metadata.yaml file is missing the following attribute(s): '{missing}'."
    )


def test_jujumetadata_missing_field_multiple(tmp_path):
    """More than one required field is missing in the metadata file."""
    # metadata file with not all fields
    metadata_file = tmp_path / "metadata.yaml"
    metadata_file.write_text(
        """
        name: foobar
    """
    )
    linter = JujuMetadata()
    result = linter.run(tmp_path)
    assert result == JujuMetadata.Result.ERRORS
    assert linter.text == (
        "The metadata.yaml file is missing the following attribute(s): "
        "'description' and 'summary'."
    )


# --- tests for analyze function


def create_fake_checker(**kwargs):
    """Create a fake Checker.

    Receive generic kwargs and process them as a dict for the defaults, as we can't declare
    the name in the function definition and then use it in the class definition.
    """
    params = {
        "check_type": "type",
        "name": "name",
        "url": "url",
        "text": "text",
        "result": "result",
    }
    params.update(kwargs)

    class FakeChecker(BaseChecker):
        check_type = params["check_type"]
        name = params["name"]
        url = params["url"]
        text = params["text"]

        def run(self, basedir):
            return params["result"]

    return FakeChecker


def test_analyze_run_everything(config):
    """Check that analyze runs all and collect the results."""
    FakeChecker1 = create_fake_checker(
        check_type=CheckType.ATTRIBUTE, name="name1", url="url1", text="text1", result="result1"
    )
    FakeChecker2 = create_fake_checker(
        check_type=CheckType.LINT, name="name2", url="url2", text="text2", result="result2"
    )
    FakeChecker3 = create_fake_checker(
        check_type=CheckType.LINT, name="returns_none", url="url3", text=None, result="result3"
    )

    # hack the first fake checker to validate that it receives the indicated path
    def dir_validator(self, basedir):
        assert basedir == pathlib.Path("test-buildpath")
        return "result1"

    FakeChecker1.run = dir_validator

    with patch("charmcraft.linters.CHECKERS", [FakeChecker1, FakeChecker2, FakeChecker3]):
        result = analyze(config, pathlib.Path("test-buildpath"))

    r1, r2, r3 = result
    assert r1.check_type == "attribute"
    assert r1.name == "name1"
    assert r1.url == "url1"
    assert r1.text == "text1"
    assert r1.result == "result1"
    assert r2.check_type == "lint"
    assert r2.name == "name2"
    assert r2.url == "url2"
    assert r2.text == "text2"
    assert r2.result == "result2"
    assert r3.name == "returns_none"
    assert r3.url == "url3"
    assert r3.text == "n/a"
    assert r3.result == "result3"


def test_analyze_ignore_attribute(config):
    """Run all checkers except the ignored attribute."""
    FakeChecker1 = create_fake_checker(
        check_type=CheckType.ATTRIBUTE,
        name="name1",
        result="res1",
        text="text1",
        url="url1",
    )
    FakeChecker2 = create_fake_checker(
        check_type=CheckType.LINT,
        name="name2",
        result="res2",
        text="text2",
        url="url2",
    )

    config.analysis.ignore.attributes.append("name1")
    with patch("charmcraft.linters.CHECKERS", [FakeChecker1, FakeChecker2]):
        result = analyze(config, pathlib.Path("somepath"))

    res1, res2 = result
    assert res1.check_type == CheckType.ATTRIBUTE
    assert res1.name == "name1"
    assert res1.result == LintResult.IGNORED
    assert res1.text == ""
    assert res1.url == "url1"
    assert res2.check_type == CheckType.LINT
    assert res2.name == "name2"
    assert res2.result == "res2"
    assert res2.text == "text2"
    assert res2.url == "url2"


def test_analyze_ignore_linter(config):
    """Run all checkers except the ignored linter."""
    FakeChecker1 = create_fake_checker(
        check_type=CheckType.ATTRIBUTE,
        name="name1",
        result="res1",
        text="text1",
        url="url1",
    )
    FakeChecker2 = create_fake_checker(
        check_type=CheckType.LINT,
        name="name2",
        result="res2",
        text="text2",
        url="url2",
    )

    config.analysis.ignore.linters.append("name2")
    with patch("charmcraft.linters.CHECKERS", [FakeChecker1, FakeChecker2]):
        result = analyze(config, pathlib.Path("somepath"))

    res1, res2 = result
    assert res1.check_type == CheckType.ATTRIBUTE
    assert res1.name == "name1"
    assert res1.result == "res1"
    assert res1.text == "text1"
    assert res1.url == "url1"
    assert res2.check_type == CheckType.LINT
    assert res2.name == "name2"
    assert res2.result == LintResult.IGNORED
    assert res2.text == ""
    assert res2.url == "url2"


def test_analyze_override_ignore(config):
    """Run all checkers even the ignored ones, if requested."""
    FakeChecker1 = create_fake_checker(check_type=CheckType.ATTRIBUTE, name="name1", result="res1")
    FakeChecker2 = create_fake_checker(check_type=CheckType.LINT, name="name2", result="res2")

    config.analysis.ignore.attributes.append("name1")
    config.analysis.ignore.linters.append("name2")
    with patch("charmcraft.linters.CHECKERS", [FakeChecker1, FakeChecker2]):
        result = analyze(config, pathlib.Path("somepath"), override_ignore_config=True)

    res1, res2 = result
    assert res1.check_type == CheckType.ATTRIBUTE
    assert res1.name == "name1"
    assert res1.result == "res1"
    assert res2.check_type == CheckType.LINT
    assert res2.name == "name2"
    assert res2.result == "res2"


def test_analyze_crash_attribute(config):
    """The attribute checker crashes."""
    FakeChecker = create_fake_checker(
        check_type=CheckType.ATTRIBUTE, name="name", text="text", url="url"
    )

    def raises(*a):
        raise ValueError

    FakeChecker.run = raises

    with patch("charmcraft.linters.CHECKERS", [FakeChecker]):
        result = analyze(config, pathlib.Path("somepath"))

    (res,) = result
    assert res.check_type == CheckType.ATTRIBUTE
    assert res.name == "name"
    assert res.result == LintResult.UNKNOWN
    assert res.text == "text"
    assert res.url == "url"


def test_analyze_crash_lint(config):
    """The lint checker crashes."""
    FakeChecker = create_fake_checker(
        check_type=CheckType.LINT, name="name", text="text", url="url"
    )

    def raises(*a):
        raise ValueError

    FakeChecker.run = raises

    with patch("charmcraft.linters.CHECKERS", [FakeChecker]):
        result = analyze(config, pathlib.Path("somepath"))

    (res,) = result
    assert res.check_type == CheckType.LINT
    assert res.name == "name"
    assert res.result == LintResult.FATAL
    assert res.text == "text"
    assert res.url == "url"


def test_analyze_all_can_be_ignored(config):
    """Control that all real life checkers can be ignored."""
    config.analysis.ignore.attributes.extend(
        c.name for c in CHECKERS if c.check_type == CheckType.ATTRIBUTE
    )
    config.analysis.ignore.linters.extend(
        c.name for c in CHECKERS if c.check_type == CheckType.LINT
    )
    result = analyze(config, pathlib.Path("somepath"))
    assert all(r.result == LintResult.IGNORED for r in result)


# --- tests for JujuActions checker


def test_jujuactions_ok(tmp_path):
    """The actions.yaml file is valid."""
    actions_file = tmp_path / "actions.yaml"
    actions_file.write_text("stuff: foobar")
    result = JujuActions().run(tmp_path)
    assert result == JujuActions.Result.OK


def test_jujuactions_missing_file(tmp_path):
    """No actions.yaml file at all."""
    result = JujuActions().run(tmp_path)
    assert result == JujuActions.Result.OK


def test_jujuactions_file_corrupted(tmp_path):
    """The actions.yaml file is not valid YAML."""
    actions_file = tmp_path / "actions.yaml"
    actions_file.write_text(" - \n-")
    result = JujuActions().run(tmp_path)
    assert result == JujuActions.Result.ERRORS


# --- tests for JujuConfig checker


def test_jujuconfig_ok(tmp_path):
    """The config.yaml file is valid."""
    config_file = tmp_path / "config.yaml"
    config_file.write_text(
        """
        options:
            foo:
                type: buzz
    """
    )
    result = JujuConfig().run(tmp_path)
    assert result == JujuConfig.Result.OK


def test_jujuconfig_missing_file(tmp_path):
    """No config.yaml file at all."""
    result = JujuConfig().run(tmp_path)
    assert result == JujuConfig.Result.OK


def test_jujuconfig_file_corrupted(tmp_path):
    """The config.yaml file is not valid YAML."""
    config_file = tmp_path / "config.yaml"
    config_file.write_text(" - \n-")
    linter = JujuConfig()
    result = linter.run(tmp_path)
    assert result == JujuConfig.Result.ERRORS
    assert linter.text == "The config.yaml file is not a valid YAML file."


def test_jujuconfig_no_options(tmp_path):
    """The config.yaml file does not have an options key."""
    config_file = tmp_path / "config.yaml"
    config_file.write_text(
        """
        summary: Small text.
    """
    )
    linter = JujuConfig()
    result = linter.run(tmp_path)
    assert result == JujuConfig.Result.ERRORS
    assert linter.text == "Error in config.yaml: must have an 'options' dictionary."


def test_jujuconfig_empty_options(tmp_path):
    """The config.yaml file has an empty options key."""
    config_file = tmp_path / "config.yaml"
    config_file.write_text(
        """
        options:
    """
    )
    linter = JujuConfig()
    result = linter.run(tmp_path)
    assert result == JujuConfig.Result.ERRORS
    assert linter.text == "Error in config.yaml: must have an 'options' dictionary."


def test_jujuconfig_options_not_dict(tmp_path):
    """The config.yaml file has an options key that is not a dict."""
    config_file = tmp_path / "config.yaml"
    config_file.write_text(
        """
        options:
          - foo
          - bar
    """
    )
    linter = JujuConfig()
    result = linter.run(tmp_path)
    assert result == JujuConfig.Result.ERRORS
    assert linter.text == "Error in config.yaml: must have an 'options' dictionary."


def test_jujuconfig_no_type_in_options_items(tmp_path):
    """The items under 'options' must have a 'type' key."""
    config_file = tmp_path / "config.yaml"
    config_file.write_text(
        """
        options:
          foo:
            description: something missing
    """
    )
    linter = JujuConfig()
    result = linter.run(tmp_path)
    assert result == JujuConfig.Result.ERRORS
    assert linter.text == "Error in config.yaml: items under 'options' must have a 'type' key."


# --- tests for Entrypoint checker


def test_entrypoint_not_used(tmp_path):
    """An entrypoint is not really used, nothing to check."""
    with patch("charmcraft.linters.get_entrypoint_from_dispatch") as mock_check:
        mock_check.return_value = None
        result = Entrypoint().run(tmp_path)
    assert result == Entrypoint.Result.NONAPPLICABLE
    mock_check.assert_called_with(tmp_path)


def test_entrypoint_all_ok(tmp_path):
    """All conditions for 'framework' are in place."""
    entrypoint = tmp_path / "charm.sh"
    entrypoint.touch(mode=0o777)
    with patch("charmcraft.linters.get_entrypoint_from_dispatch") as mock_check:
        mock_check.return_value = entrypoint
        result = Entrypoint().run(tmp_path)
    assert result == Entrypoint.Result.OK
    mock_check.assert_called_with(tmp_path)


def test_entrypoint_missing(tmp_path):
    """The file does not exist."""
    entrypoint = tmp_path / "charm"
    linter = Entrypoint()
    with patch("charmcraft.linters.get_entrypoint_from_dispatch", return_value=entrypoint):
        result = linter.run(tmp_path)
    assert result == Entrypoint.Result.ERRORS
    assert linter.text == f"Cannot find the entrypoint file: {str(entrypoint)!r}"


def test_entrypoint_directory(tmp_path):
    """The pointed entrypoint is a directory."""
    entrypoint = tmp_path / "charm"
    entrypoint.mkdir()
    linter = Entrypoint()
    with patch("charmcraft.linters.get_entrypoint_from_dispatch", return_value=entrypoint):
        result = linter.run(tmp_path)
    assert result == Entrypoint.Result.ERRORS
    assert linter.text == f"The entrypoint is not a file: {str(entrypoint)!r}"


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_entrypoint_non_exec(tmp_path):
    """The pointed entrypoint is not an executable file."""
    entrypoint = tmp_path / "charm"
    entrypoint.touch()
    linter = Entrypoint()
    with patch("charmcraft.linters.get_entrypoint_from_dispatch", return_value=entrypoint):
        result = linter.run(tmp_path)
    assert result == Entrypoint.Result.ERRORS
    assert linter.text == f"The entrypoint file is not executable: {str(entrypoint)!r}"



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Reformat pydantic errors."""

from typing import Tuple


def format_pydantic_error_location(loc) -> str:
    """Format location."""
    loc_parts = []
    for loc_part in loc:
        if isinstance(loc_part, str):
            loc_parts.append(loc_part)
        elif isinstance(loc_part, int):
            # Integer indicates an index. Go
            # back and fix up previous part.
            previous_part = loc_parts.pop()
            previous_part += f"[{loc_part}]"
            loc_parts.append(previous_part)
        else:
            raise RuntimeError(f"unhandled loc: {loc_part}")

    loc = ".".join(loc_parts)

    # Filter out internal __root__ detail.
    return loc.replace(".__root__", "")


def format_pydantic_error_message(msg: str) -> str:
    """Format pydantic's error message field."""
    # Replace shorthand "str" with "string".
    return msg.replace("str type expected", "string type expected")


def printable_field_location_split(location: str) -> Tuple[str, str]:
    """Return split field location.

    If top-level, location is returned as unquoted "top-level".
    If not top-level, location is returned as quoted location.

    Examples
    --------
    (1) field1[idx].foo => 'foo', 'field1[idx]'
    (2) field2 => 'field2', top-level

    :returns: Tuple of <field name>, <location> as printable representations.

    """
    loc_split = location.split(".")
    field_name = repr(loc_split.pop())

    if loc_split:
        return field_name, repr(".".join(loc_split))

    return field_name, "top-level"


def format_pydantic_errors(errors, *, file_name: str = "charmcraft.yaml") -> str:
    """Format errors.

    Example 1: Single error.

    Bad charmcraft.yaml content:
    - field: <some field>
      reason: <some reason>

    Example 2: Multiple errors.

    Bad charmcraft.yaml content:
    - field: <some field>
      reason: <some reason>
    - field: <some field 2>
      reason: <some reason 2>
    """
    combined = [f"Bad {file_name} content:"]
    for error in errors:
        formatted_loc = format_pydantic_error_location(error["loc"])
        formatted_msg = format_pydantic_error_message(error["msg"])

        if formatted_msg == "field required":
            field_name, location = printable_field_location_split(formatted_loc)
            combined.append(f"- field {field_name} required in {location} configuration")
        elif formatted_msg == "extra fields not permitted":
            field_name, location = printable_field_location_split(formatted_loc)
            combined.append(
                f"- extra field {field_name} not permitted in {location} configuration"
            )
        else:
            combined.append(f"- {formatted_msg} in field {formatted_loc!r}")

    return "\n".join(combined)
